#!/usr/bin/python
# service_message_history.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_message_history.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_message_history
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return MessageHistoryService()


class MessageHistoryService(LocalService):

    service_name = 'service_message_history'
    config_path = 'services/message-history/enabled'

    def dependent_on(self):
        return [
            'service_my_data',
            'service_private_messages',
        ]

    def start(self):
        from bitdust.chat import message_database
        from bitdust.chat import message_keeper
        from bitdust.main import events
        from bitdust.main import listeners
        message_database.init()
        message_keeper.init()
        events.add_subscriber(self.on_key_registered, 'key-registered')
        events.add_subscriber(self.on_key_renamed, 'key-renamed')
        events.add_subscriber(self.on_key_generated, 'key-generated')
        events.add_subscriber(self.on_key_erased, 'key-erased')
        if listeners.is_populate_required('conversation'):
            # listeners.populate_later().remove('conversation')
            message_database.populate_conversations()
        if listeners.is_populate_required('message'):
            # listeners.populate_later().remove('message')
            message_database.populate_messages()
        return True

    def stop(self):
        from bitdust.chat import message_database
        from bitdust.chat import message_keeper
        from bitdust.main import events
        events.remove_subscriber(self.on_key_erased, 'key-erased')
        events.remove_subscriber(self.on_key_generated, 'key-generated')
        events.remove_subscriber(self.on_key_renamed, 'key-renamed')
        events.remove_subscriber(self.on_key_registered, 'key-registered')
        message_keeper.shutdown()
        message_database.shutdown()
        return True

    def health_check(self):
        return True

    def on_key_generated(self, evt):
        self.do_check_create_rename_key(evt.data['key_id'])

    def on_key_registered(self, evt):
        self.do_check_create_rename_key(evt.data['key_id'])

    def on_key_renamed(self, evt):
        self.do_check_create_rename_key(evt.data['new_key_id'])

    def on_key_erased(self, evt):
        from bitdust.main import listeners
        from bitdust.chat import message_database
        if evt.data['key_id'].startswith('group_'):
            conversation_id = message_database.get_conversation_id(evt.data['local_key_id'], evt.data['local_key_id'], 3)
            listeners.push_snapshot('conversation', snap_id=conversation_id, deleted=True)

    def do_check_create_rename_key(self, new_key_id):
        from bitdust.logs import lg
        from bitdust.crypt import my_keys
        from bitdust.chat import message_database
        try:
            new_public_key = my_keys.get_public_key_raw(new_key_id)
        except:
            lg.exc()
            return
        try:
            new_local_key_id = my_keys.get_local_key_id(new_key_id)
        except:
            lg.exc()
            return
        if new_local_key_id is None:
            lg.err('did not found local_key_id for %r' % new_key_id)
            return
        message_database.check_create_rename_key(
            new_public_key=new_public_key,
            new_key_id=new_key_id,
            new_local_key_id=new_local_key_id,
        )



#!/usr/bin/python
# backup_matrix.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (backup_matrix.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: backup_matrix.

The software stores all backup IDs in the memory.
Also we need to keep track of every piece of data.
I have a matrix for every backup: blocks by suppliers.
Every cell in the matrix store short info about single piece (file) of your data.

There are two types of matrix:
1. "remote files"
2. "local files"

A "remote" matrix keeps info about files stored on your suppliers.
They will report this info to you in the command "ListFiles".

A "local" matrix keeps info about local files stored on your machine.
When you doing a backup you create a two copies of your data.
At first it will be stored on your local HDD and then transferred to suppliers.

The local and remote copies of your data is absolutely equal, have same structure and dimension.

Local files can be removed as soon as corresponding remote files gets delivered to suppliers.
But local files is needed to rebuild the data - the "Parity" pieces is used in the RAID code
to reconstruct "Data" pieces. So need to keep track of both "surfaces".
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from six.moves import range
from io import BytesIO

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import os

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio

from bitdust.lib import packetid
from bitdust.lib import misc
from bitdust.lib import strng

from bitdust.main import settings
from bitdust.main import listeners

from bitdust.contacts import contactsdb

from bitdust.services import driver

from bitdust.crypt import my_keys

from bitdust.storage import backup_fs

from bitdust.userid import my_id
from bitdust.userid import global_id
from bitdust.userid import id_url

#------------------------------------------------------------------------------

_RemoteFiles = {}
_LocalFiles = {}
_RemoteMaxBlockNumbers = {}
_LocalMaxBlockNumbers = {}
_LocalBackupSize = {}
_BackupStatusNotifyCallback = None
_StatusCallBackForGuiBackup = None
_LocalFilesNotifyCallback = None
_UpdatedBackupIDs = set()
_ListFilesQueryCallbacks = {}

#------------------------------------------------------------------------------


def init():
    """
    Call this method before all others here. Prepare several things here:

    * start a loop to repaint the GUI when needed
    * scan local files and build the "local" matrix
    * read latest (stored on local disk) ListFiles for suppliers to build "remote" matrix
    """
    if _Debug:
        lg.out(_DebugLevel, 'backup_matrix.init')
    ReadLocalFiles()
    ReadLatestRawListFiles()


def shutdown():
    """
    Correct way to finish all things here.
    """
    if _Debug:
        lg.out(_DebugLevel, 'backup_matrix.shutdown')


#------------------------------------------------------------------------------


def remote_files():
    """
    This is a "remote" matrix. Here is stored info for all remote files (on
    suppliers HDD's) stored in dictionary. The values are integers::

      -1 : this mean file is missing
      0  : no info comes yet
      1  : this file exist on given remote machine

    This is a dictionary of dictionaries of dictionaries of lists. :-)))
    Values can be accessed this way::

      remote_files()[backupID][blockNumber][dataORparity][supplierNumber]

    Here the keys are:

    - backupID - a unique identifier of that backup, see ``lib.packetid`` module
    - blockNumber - a number of block started from 0, look at ``p2p.backup``
    - dataORparity - can be 'D' for Data packet or 'P' for Parity packets
    - supplierNumber - who should keep that piece?
    """
    global _RemoteFiles
    return _RemoteFiles


def remote_max_block_numbers():
    """
    This is a dictionary to store max block number for every remote backup, by
    backup ID.
    """
    global _RemoteMaxBlockNumbers
    return _RemoteMaxBlockNumbers


def local_files():
    """
    This is a "local" matrix, same structure like ``remote_files()``.

    Keeps info for all local files stored on your HDD. The values are
    integers: 0 or 1 to know that local file exists or not.
    """
    global _LocalFiles
    return _LocalFiles


def local_max_block_numbers():
    """
    Dictionary to store max block number for every local backup.
    """
    global _LocalMaxBlockNumbers
    return _LocalMaxBlockNumbers


def local_backup_size():
    """
    The pieces can have different sizes.

    To get the size of the particular backup need to count size of all
    pieces. Here is a dictionary of counters for every backup.
    """
    global _LocalBackupSize
    return _LocalBackupSize


#------------------------------------------------------------------------------


def GetActiveArray(customer_idurl=None):
    """
    Loops all suppliers and returns who is alive at the moment.

    Return a list with integers: 0 for offline suppler and 1 if he is
    available right now. Uses ``p2p.online_status.isOnline()`` to see
    the current state of supplier.
    """
    from bitdust.p2p import online_status
    activeArray = [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl)
    for i in range(contactsdb.num_suppliers(customer_idurl=customer_idurl)):
        suplier_idurl = contactsdb.supplier(i, customer_idurl=customer_idurl)
        if not suplier_idurl:
            activeArray[i] = 0
            continue
        if online_status.isOnline(suplier_idurl):
            activeArray[i] = 1
        else:
            activeArray[i] = 0
    return activeArray


def SuppliersChangedNumbers(oldSupplierList, customer_idurl=None):
    """
    Return list of positions of changed suppliers, say if suppliers 1 and 3
    were changed it should return [1,3].
    """
    changedList = []
    for i in range(len(oldSupplierList)):
        old_supplier_idurl = oldSupplierList[i]
        new_supplier_idurl = contactsdb.supplier(i, customer_idurl=customer_idurl)
        if not old_supplier_idurl:
            if new_supplier_idurl:
                changedList.append(i)
                lg.info('found empty supplier on position %d replaced with new supplier %r' % (i, new_supplier_idurl))
            else:
                lg.warn('found empty supplier on position %d which was not replaced yet' % i)
            continue
        if not id_url.is_the_same(new_supplier_idurl, old_supplier_idurl):
            changedList.append(i)
            lg.info('found supplier change on position %d: %r -> %r' % (i, old_supplier_idurl, new_supplier_idurl))
    if _Debug:
        lg.args(_DebugLevel, changed=changedList, old_suppliers=oldSupplierList)
    return changedList


#------------------------------------------------------------------------------


def DetectSupplierPosition(raw_list_file_text):
    inpt = BytesIO(strng.to_bin(raw_list_file_text))
    all_positions = {}
    while True:
        line = strng.to_text(inpt.readline())
        if line == '':
            break
        typ = line[0]
        line = line[1:]
        line = line.rstrip('\n')
        if line.strip() == '':
            continue
        if line.find('http://') != -1 or line.find('.xml') != -1:
            continue
        if typ == 'V':
            # minimum is 4 words: "0/0/F20090709034221PM", "3", "0-1000" "123456"
            words = line.split(' ')
            if len(words) < 4:
                lg.warn('incorrect line (words count): [%s]' % line)
                continue
            try:
                supplier_pos = int(words[1])
            except:
                lg.warn('incorrect line: [%s]' % line)
                continue
            if supplier_pos not in all_positions:
                all_positions[supplier_pos] = 0
            all_positions[supplier_pos] += 1
    inpt.close()
    if _Debug:
        lg.out(_DebugLevel, 'backup_matrix.DetectSupplierPosition from %d bytes found: %s' % (len(raw_list_file_text), all_positions))
    if not all_positions:
        return -1
    all_positions = list(all_positions.items())
    all_positions.sort(key=lambda i: i[1], reverse=True)
    return all_positions[0][0]


#------------------------------------------------------------------------------


def process_line_key(line):
    return line.strip()


def process_line_dir(line, current_key_alias=None, customer_idurl=None, is_in_sync=None, ignored_path_ids=[], auto_create=False):
    paths2remove = set()
    modified = False
    try:
        pth = line.split(' ')[0]
    except:
        pth = line
    path_id = pth.strip('/')
    if auto_create and is_in_sync:
        if path_id != settings.BackupIndexFileName() and path_id not in ignored_path_ids:
            if not backup_fs.ExistsID(pth, iterID=backup_fs.fsID(customer_idurl, current_key_alias)):
                if _Debug:
                    lg.out(_DebugLevel, '        AUTO CREATE DIR "%s" in the index' % pth)
                item = backup_fs.FSItemInfo(
                    name=path_id,
                    path_id=path_id,
                    typ=backup_fs.DIR,
                    key_id=global_id.MakeGlobalID(idurl=customer_idurl, key_alias=current_key_alias) if current_key_alias else None,
                )
                success, _modified = backup_fs.SetDir(item, customer_idurl=customer_idurl)
                if _modified:
                    modified = True
    if not backup_fs.ExistsID(pth, iterID=backup_fs.fsID(customer_idurl, current_key_alias)):
        if is_in_sync:
            # if customer_idurl == my_id.getIDURL():
            paths2remove.add(packetid.MakeBackupID(
                customer=global_id.UrlToGlobalID(customer_idurl),
                path_id=pth,
                key_alias=current_key_alias,
                version=None,
            ))
            if _Debug:
                lg.out(_DebugLevel, '        DIR "%s" to be removed, not found in the index' % pth)
        # else:
        #     if _Debug:
        #         lg.out(_DebugLevel, '        DIR "%s" skip removing, index not in sync' % pth)
    return modified, paths2remove


def process_line_file(line, current_key_alias=None, customer_idurl=None, is_in_sync=None, ignored_path_ids=[], auto_create=False):
    """
    if we don't have this path in the index at the moment we have several possible scenarios:
       1. this is old file and we need to remove it and all its backups
       2. we loose our local index and did not restore it from one of suppliers yet
       3. we did restore our account but did not restore the index yet
       4. we lost our index at all and we do not have nor local nor remote copy
       5. this is a file stored by supplier who holds data for another customer
       6. this is my own file, but uploaded directly to the supplier by third node possessing the public key
    what to do now:
       - in first case we just need to remove the file from remote supplier
       - we must not remove the data if it is not my customer or file uploaded directly to the supplier from external
       - in other cases we must keep all remote data and believe we can restore the index
            and get all file names and backed up data
    how to recognize that? how to be sure we have the correct index?
    because it should be empty right after we recover our account
    or we may loose it if the local index file were lost
    the first idea:  check index_synchronizer() state - IN_SYNC means index is fine
    the second idea: check revision number of the local index - 0 means we have no index yet
    """
    paths2remove = set()
    modified = False
    try:
        pth, filesz = line.split(' ')
        filesz = int(filesz)
    except:
        pth = line
        filesz = -1
    path_id = pth.strip('/')
    if auto_create and is_in_sync:
        if path_id != settings.BackupIndexFileName() and path_id not in ignored_path_ids:
            if not backup_fs.IsFileID(pth, iterID=backup_fs.fsID(customer_idurl, current_key_alias)):
                if _Debug:
                    lg.out(_DebugLevel, '        AUTO CREATE FILE "%s" in the index' % pth)
                item = backup_fs.FSItemInfo(
                    name=path_id,
                    path_id=path_id,
                    typ=backup_fs.FILE,
                    key_id=global_id.MakeGlobalID(idurl=customer_idurl, key_alias=current_key_alias) if current_key_alias else None,
                )
                item.size = filesz
                success, _modified = backup_fs.SetFile(item, customer_idurl=customer_idurl)
                if _modified:
                    modified = True
    if not backup_fs.IsFileID(pth, iterID=backup_fs.fsID(customer_idurl, current_key_alias)):
        # remote supplier have some file - but we don't have it in the index
        if path_id == settings.BackupIndexFileName():
            # this is the index file saved on remote supplier
            # must remember its size and put it in the backup_fs
            item = backup_fs.FSItemInfo(
                name=path_id,
                path_id=path_id,
                typ=backup_fs.FILE,
                key_id=global_id.MakeGlobalID(idurl=customer_idurl, key_alias=current_key_alias),
            )
            item.size = filesz
            success, _modified = backup_fs.SetFile(item, customer_idurl=customer_idurl)
            if _modified:
                modified = True
        else:
            if is_in_sync:
                # so we have some modifications in the index - it is not empty!
                # index_synchronizer() did the job - so we have up to date index on hands
                # if customer_idurl == my_id.getIDURL():
                # now we are sure that this file is old and must be removed from remote supplier
                paths2remove.add(packetid.MakeBackupID(
                    customer=global_id.UrlToGlobalID(customer_idurl),
                    path_id=pth,
                    key_alias=current_key_alias,
                    version=None,
                ))
                if _Debug:
                    lg.out(_DebugLevel, '        FILE "%s" to be removed, not found in the index' % pth)
            else:
                if _Debug:
                    lg.out(_DebugLevel, '        FILE "%s" skip removing, index not in sync yet' % pth)
                # what to do now? let's hope we still can restore our index and this file is our remote data
    return modified, paths2remove


def process_line_version(line, supplier_num, current_key_alias=None, customer_idurl=None, is_in_sync=None, ignored_path_ids=[], auto_create=False):
    backups2remove = set()
    paths2remove = set()
    found_backups = set()
    newfiles = 0
    modified = False
    file_auto_created = False
    # minimum is 4 words: "0/0/F20090709034221PM", "3", "0-1000" "123456"
    words = line.split(' ')
    if len(words) < 4:
        lg.err('incorrect line (words count): [%s]' % line)
        return modified, backups2remove, paths2remove, found_backups, newfiles
    try:
        _, remotePath, versionName = packetid.SplitBackupID(words[0])
        backupID = packetid.MakeBackupID(
            customer=global_id.UrlToGlobalID(customer_idurl),
            path_id=remotePath,
            key_alias=current_key_alias,
            version=versionName,
        )
    except:
        lg.err('incorrect line (global id format): [%s]' % line)
        return modified, backups2remove, paths2remove, found_backups, newfiles
    try:
        lineSupplierNum = int(words[1])
        _, maxBlockNum = words[2].split('-')
        maxBlockNum = int(maxBlockNum)
        versionSize = int(words[3])
    except:
        lg.err('incorrect line (digits format): [%s]' % line)
        return modified, backups2remove, paths2remove, found_backups, newfiles
    if remotePath in ignored_path_ids:
        # this mean supplier have old files and we do not need those files
        # if not id_url.is_the_same(customer_idurl, my_id.getIDURL()):
        #     if _Debug:
        #         lg.out(_DebugLevel, '        VERSION "%s" skip removing, found unknown data from another customer: %r' % (backupID, customer_idurl))
        #     return modified, backups2remove, paths2remove, found_backups, newfiles
        backups2remove.add(backupID)
        paths2remove.add(packetid.MakeBackupID(
            customer=global_id.UrlToGlobalID(customer_idurl),
            path_id=remotePath,
            key_alias=current_key_alias,
            version=None,
        ))
        if _Debug:
            lg.out(_DebugLevel, '        VERSION "%s" to be removed, ignoring path %s because it was deleted' % (backupID, remotePath))
        return modified, backups2remove, paths2remove, found_backups, newfiles
    if lineSupplierNum != supplier_num:
        # if not id_url.is_the_same(customer_idurl, my_id.getIDURL()):
        #     if _Debug:
        #         lg.out(_DebugLevel, '        VERSION "%s" skip removing, found unknown data from another customer: %r' % (backupID, customer_idurl))
        #     return modified, backups2remove, paths2remove, found_backups, newfiles
        # this mean supplier have old files and we do not need those files
        backups2remove.add(backupID)
        if _Debug:
            lg.out(_DebugLevel, '        VERSION "%s" to be removed, different supplier number' % backupID)
        return modified, backups2remove, paths2remove, found_backups, newfiles
    iter_path = backup_fs.WalkByID(remotePath, iterID=backup_fs.fsID(customer_idurl, current_key_alias))
    item = None
    if iter_path:
        item = iter_path[0]
    if isinstance(item, dict):
        try:
            item = item[backup_fs.INFO_KEY]
        except:
            item = None
    if item is None:
        # this path is not found in the index at all
        if is_in_sync:
            # if not id_url.is_the_same(customer_idurl, my_id.getIDURL()):
            #     if _Debug:
            #         lg.out(_DebugLevel, '        VERSION "%s" skip removing, found unknown data from another customer: %r' % (backupID, customer_idurl))
            #     return modified, backups2remove, paths2remove, found_backups, newfiles
            found_backups.add(backupID)
            backups2remove.add(backupID)
            paths2remove.add(packetid.MakeBackupID(
                customer=global_id.UrlToGlobalID(customer_idurl),
                path_id=remotePath,
                key_alias=current_key_alias,
                version=None,
            ))
            if _Debug:
                lg.out(_DebugLevel, '        VERSION "%s" to be removed, path not found in the index' % backupID)
        else:
            if _Debug:
                lg.out(_DebugLevel, '        VERSION "%s" skip removing, index not in sync' % backupID)
        return modified, backups2remove, paths2remove, found_backups, newfiles
    if auto_create and is_in_sync:
        if not item.has_version(versionName):
            if not current_key_alias or not customer_idurl:
                if _Debug:
                    lg.out(_DebugLevel, '        AUTO CREATE VERSION (skip key verification) "%s" at "%s" in the index' % (versionName, remotePath))
                item.add_version(versionName)
                modified = True
                file_auto_created = True
            else:
                authorized_key_id = my_keys.make_key_id(
                    alias=current_key_alias,
                    creator_idurl=customer_idurl,
                )
                if my_keys.is_key_registered(authorized_key_id):
                    if _Debug:
                        lg.out(_DebugLevel, '        AUTO CREATE VERSION "%s" at "%s" in the index' % (versionName, remotePath))
                    item.add_version(versionName)
                    modified = True
                    file_auto_created = True
                else:
                    lg.warn('skip auto create version %r for path %r because key %r not registered' % (versionName, remotePath, authorized_key_id))
    if not item.has_version(versionName):
        if is_in_sync:
            # if not id_url.is_the_same(customer_idurl, my_id.getIDURL()):
            #     if _Debug:
            #         lg.out(_DebugLevel, '        VERSION "%s" skip removing, found unknown data from another customer: %r' % (backupID, customer_idurl))
            #     return modified, backups2remove, paths2remove, found_backups, newfiles
            backups2remove.add(backupID)
            if _Debug:
                lg.out(_DebugLevel, '        VERSION "%s" to be removed, version is not found in the index' % backupID)
        else:
            if _Debug:
                lg.out(_DebugLevel, '        VERSION "%s" skip removing, index not in sync' % backupID)
        return modified, backups2remove, paths2remove, found_backups, newfiles
    else:
        if _Debug:
            lg.out(_DebugLevel, '        VERSION "%s" is found in the index' % backupID)
    item_version_info = item.get_version_info(versionName)
    missingBlocksSet = {'Data': set(), 'Parity': set()}
    if len(words) > 4:
        # "0/0/123/4567/F20090709034221PM/0-Data" "3" "0-5" "434353" "missing" "Data:1,3" "Parity:0,1,2"
        if words[4].strip() != 'missing':
            lg.err('incorrect line:[%s]' % line)
            return modified, backups2remove, paths2remove, found_backups, newfiles
        for missingBlocksString in words[5:]:
            try:
                dp, blocks = missingBlocksString.split(':')
                missingBlocksSet[dp] = set(blocks.split(','))
            except:
                lg.exc()
                return modified, backups2remove, paths2remove, found_backups, newfiles
    if backupID not in remote_files():
        remote_files()[backupID] = {}
        if _Debug:
            lg.out(_DebugLevel, '            new remote entry for %s created in memory' % backupID)
    # +1 because range(2) give us [0,1] but we want [0,1,2]
    for blockNum in range(maxBlockNum + 1):
        if blockNum not in remote_files()[backupID]:
            remote_files()[backupID][blockNum] = {
                'D': [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl),
                'P': [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl),
            }
        for dataORparity in ['Data', 'Parity']:
            # we set -1 if the file is missing and 1 if exist, so 0 mean "no info yet" ... smart!
            bit = -1 if str(blockNum) in missingBlocksSet[dataORparity] else 1
            remote_files()[backupID][blockNum][dataORparity[0]][supplier_num] = bit
            newfiles += int((bit + 1)/2)  # this should switch -1 or 1 to 0 or 1
    # save max block number for this backup
    if backupID not in remote_max_block_numbers():
        remote_max_block_numbers()[backupID] = -1
    if maxBlockNum > remote_max_block_numbers()[backupID]:
        remote_max_block_numbers()[backupID] = maxBlockNum
    if len(missingBlocksSet['Data']) == 0 and len(missingBlocksSet['Parity']) == 0:
        found_backups.add(backupID)
    if item_version_info[0] != maxBlockNum or (item_version_info[1] in [None, -1, 0] and versionSize > 0):
        if _Debug:
            lg.out(_DebugLevel, '            updating version %s info, maxBlockNum %r->%r, size %r->%r' % (
                backupID,
                item_version_info[0],
                maxBlockNum,
                item_version_info[1],
                versionSize,
            ))
        item.set_version_info(versionName, maxBlockNum, versionSize)
        modified = True
    if file_auto_created or newfiles:
        full_remote_path = global_id.MakeGlobalID(path=item.name(), key_id=item.key_id)
        full_remote_path_id = global_id.MakeGlobalID(path=item.path_id, key_id=item.key_id)
        _, percent, _, weakPercent = GetBackupRemoteStats(backupID)
        snapshot = dict(
            backup_id=backupID,
            max_block=maxBlockNum,
            remote_path=full_remote_path,
            global_id=full_remote_path_id,
            type=item.type,
            filesize=item.size,
            size=versionSize,
            key_id=item.key_id,
            delivered=misc.percent2string(percent),
            reliable=misc.percent2string(weakPercent),
        )
        listeners.push_snapshot('remote_version', snap_id=backupID, data=snapshot)
    return modified, backups2remove, paths2remove, found_backups, newfiles


def process_raw_list_files(supplier_num, list_files_text_body, customer_idurl=None, is_in_sync=None):
    """
    Read ListFiles packet for given supplier and build a "remote" matrix. All
    lines are something like that:

      Q*
      Kmaster
      Findex 5456
      D0 -1
      D0/1 -1
      V0/1/F20090709034221PM 3 0-1000 7463434
      D0/0/123/4567 -1
      V0/0/123/4567/F20090709034221PM 3 0-11 434353 missing Data:1,3
      V0/0/123/4/F20090709012331PM 3 0-5 434353 missing Data:1,3 Parity:0,1,2
      Kkey_abc_123
      Findex 205
      D0 -1
      D0/1 -1
      V0/1/F20090709010203PM 3 0-3 174634

    First character can be:

      "Q" for selection query
      "K" for keys
      "D" for folders
      "F" for files
      "V" for stored data
    """
    global _ListFilesQueryCallbacks
    from bitdust.storage import backup_control
    if not customer_idurl:
        customer_idurl = my_id.getIDURL()
    if _Debug:
        lg.out(_DebugLevel, 'backup_matrix.process_raw_list_files [%d] : %d bytes, is_in_sync=%s, rev:%d, c=%s' % (supplier_num, len(list_files_text_body), is_in_sync, backup_fs.revision(), customer_idurl))
    backups2remove = set()
    paths2remove = set()
    missed_backups = set(remote_files().keys())
    oldfiles = 0
    newfiles = 0
    remote_files_changed = False
    current_key_alias = 'master'
    current_query = None
    current_ignored_path_ids = set()
    query_results = set()
    updated_keys = []
    inpt = BytesIO(strng.to_bin(list_files_text_body))
    while True:
        line = strng.to_text(inpt.readline())
        if line == '':
            break
        typ = line[0]
        line = line[1:]
        line = line.rstrip('\n')
        if line.strip() == '':
            continue
        # also don't consider the identity a backup
        if line.find('http://') != -1 or line.find('.xml') != -1:
            continue

        if typ == 'Q':
            current_query = line.strip()
            if _Debug:
                lg.out(_DebugLevel, '    %s %s' % (typ, current_query))
            continue

        if typ == 'K':
            current_key_alias = process_line_key(line)
            current_ignored_path_ids.clear()
            if current_key_alias != 'master':
                if driver.is_on('service_shared_data'):
                    from bitdust.access import shared_access_coordinator
                    current_ignored_path_ids.update(shared_access_coordinator.get_deleted_path_ids(customer_idurl, current_key_alias))
            if _Debug:
                lg.out(_DebugLevel, '    %s %s/%s' % (typ, current_query, current_key_alias))
            oldfiles += ClearSupplierRemoteInfo(supplier_num, customer_idurl=customer_idurl, key_alias=current_key_alias)
            continue

        if typ == 'D':
            if current_key_alias == 'master' and not id_url.is_the_same(customer_idurl, my_id.getIDURL()):
                if _Debug:
                    lg.out(_DebugLevel, '    %s %s/%s/%s IGNORED' % (typ, current_query, current_key_alias, line))
                continue
            modified, _paths2remove = process_line_dir(
                line,
                current_key_alias=current_key_alias,
                customer_idurl=customer_idurl,
                is_in_sync=is_in_sync,
                ignored_path_ids=current_ignored_path_ids,
                auto_create=False,
            )
            paths2remove.update(_paths2remove)
            remote_files_changed = remote_files_changed or modified
            if modified:
                updated_keys.append(current_key_alias)
            if _Debug:
                lg.out(_DebugLevel, '    %s %s/%s/%s %s' % (typ, current_query, current_key_alias, line, 'MODIFIED' if modified else 'IN_SYNC'))
            continue

        if typ == 'F':
            if current_key_alias == 'master' and not id_url.is_the_same(customer_idurl, my_id.getIDURL()):
                if _Debug:
                    lg.out(_DebugLevel, '    %s %s/%s/%s IGNORED' % (typ, current_query, current_key_alias, line))
                continue
            modified, _paths2remove = process_line_file(
                line,
                current_key_alias=current_key_alias,
                customer_idurl=customer_idurl,
                is_in_sync=is_in_sync,
                ignored_path_ids=current_ignored_path_ids,
                auto_create=False,
            )
            paths2remove.update(_paths2remove)
            remote_files_changed = remote_files_changed or modified
            if modified:
                updated_keys.append(current_key_alias)
            if _Debug:
                lg.out(_DebugLevel, '    %s %s/%s/%s %s' % (typ, current_query, current_key_alias, line, 'MODIFIED' if modified else 'IN_SYNC'))
            continue

        if typ == 'V':
            if current_key_alias == 'master' and not id_url.is_the_same(customer_idurl, my_id.getIDURL()):
                if _Debug:
                    lg.out(_DebugLevel, '    %s %s/%s/%s IGNORED' % (typ, current_query, current_key_alias, line))
                continue
            modified, _backups2remove, _paths2remove, found_backups, _newfiles = process_line_version(
                line,
                supplier_num=supplier_num,
                current_key_alias=current_key_alias,
                customer_idurl=customer_idurl,
                is_in_sync=is_in_sync,
                ignored_path_ids=current_ignored_path_ids,
                auto_create=False,
            )
            backups2remove.update(_backups2remove)
            paths2remove.update(_paths2remove)
            missed_backups.difference_update(found_backups)
            newfiles += _newfiles
            remote_files_changed = remote_files_changed or modified
            if current_query is not None:
                query_results.add((customer_idurl, current_query))
            if modified:
                updated_keys.append(current_key_alias)
            if _Debug:
                lg.out(_DebugLevel, '    %s %s/%s/%s %s' % (typ, current_query, current_key_alias, line, 'MODIFIED' if modified else 'IN_SYNC'))
            continue

        raise Exception('unexpected line received: %r' % line)

    inpt.close()
    remote_files_changed = remote_files_changed or (oldfiles != newfiles)
    if _Debug:
        lg.out(
            _DebugLevel, 'backup_matrix.process_raw_list_files   remote_files_changed:%s old:%d new:%d backups2remove:%d paths2remove:%d missed_backups:%d remote_files:%d query_results:%d' % (
                remote_files_changed,
                oldfiles,
                newfiles,
                len(backups2remove),
                len(paths2remove),
                len(missed_backups),
                len(remote_files()),
                len(query_results),
            )
        )
    if remote_files_changed and is_in_sync:
        for key_alias in updated_keys:
            backup_control.SaveFSIndex(customer_idurl, key_alias)
    for query_key in query_results:
        if query_key in _ListFilesQueryCallbacks:
            for cb in _ListFilesQueryCallbacks[query_key]:
                cb(supplier_num, newfiles)
    query_results.clear()
    # finally return list of items which are too old but stored on suppliers machines
    return remote_files_changed, backups2remove, paths2remove, missed_backups


#------------------------------------------------------------------------------


def SaveLatestRawListFiles(supplier_idurl, raw_data, customer_idurl=None):
    """
    Save a ListFiles packet from given supplier on local HDD.
    """
    if not customer_idurl:
        customer_idurl = my_id.getIDURL()
    if _Debug:
        lg.out(_DebugLevel, 'backup_matrix.SaveLatestRawListFiles, %s, customer_idurl=%s' % (supplier_idurl, customer_idurl))
    supplierPath = settings.SupplierPath(supplier_idurl, customer_idurl)
    if not os.path.isdir(supplierPath):
        try:
            os.makedirs(supplierPath)
        except:
            lg.exc()
            return
    bpio.WriteTextFile(settings.SupplierListFilesFilename(supplier_idurl, customer_idurl), raw_data)


def ReadLatestRawListFiles(customer_idurl=None):
    """
    Call ``process_raw_list_files()`` for every local file we have on hands and build
    whole "remote" matrix.
    """
    if not customer_idurl:
        customer_idurl = my_id.getIDURL()
    if _Debug:
        lg.out(_DebugLevel, 'backup_matrix.ReadLatestRawListFiles  customer_idurl=%r' % customer_idurl)
    for idurl in contactsdb.suppliers(customer_idurl=customer_idurl):
        if idurl:
            filename = os.path.join(settings.SupplierPath(idurl, customer_idurl, 'listfiles'))
            if os.path.isfile(filename):
                listFileText = bpio.ReadTextFile(filename).strip()
                if listFileText:
                    remote_files_changed, backups2remove, paths2remove, missed_backups = process_raw_list_files(
                        supplier_num=contactsdb.supplier_position(idurl),
                        list_files_text_body=listFileText,
                        customer_idurl=customer_idurl,
                        is_in_sync=False,
                    )
                    if _Debug:
                        lg.out(
                            _DebugLevel,
                            '    %r loaded with %d bytes, changed:%r, backups2remove:%d, paths2remove:%d, missed_backups:%d' % (filename, len(listFileText), remote_files_changed, len(backups2remove), len(paths2remove), len(missed_backups))
                        )


#------------------------------------------------------------------------------


def ReadLocalFiles():
    """
    This method scans local backups and build the whole "local" matrix.
    """
    global _LocalFilesNotifyCallback
    local_files().clear()
    local_max_block_numbers().clear()
    local_backup_size().clear()
    _counter = [
        0,
    ]

    def visit(key_id, realpath, subpath, name):
        # subpath is something like 0/0/1/0/F20131120053803PM/0-1-Data
        if not os.path.isfile(realpath):
            return True
        if realpath.startswith('newblock-'):
            return False
        if subpath == settings.BackupIndexFileName():
            return False
        try:
            version = subpath.split('/')[-2]
        except:
            return False
        if not packetid.IsCanonicalVersion(version):
            return True
        LocalFileReport(packetID=packetid.MakeBackupID(key_id, subpath))
        _counter[0] += 1
        return False

    all_keys = os.listdir(settings.getLocalBackupsDir())
    for key_id in all_keys:
        latest_key_id = my_keys.latest_key_id(key_id)
        if key_id != latest_key_id:
            old_path = os.path.join(settings.getLocalBackupsDir(), key_id)
            new_path = os.path.join(settings.getLocalBackupsDir(), my_keys.latest_key_id(key_id))
            if os.path.isdir(old_path):
                try:
                    bpio.move_dir_recursive(old_path, new_path)
                    if _Debug:
                        lg.dbg(_DebugLevel, 'copied %r into %r' % (old_path, new_path))
                    if os.path.exists(old_path):
                        bpio._dir_remove(old_path)
                        if _Debug:
                            lg.dbg(_DebugLevel, 'removed %r' % old_path)
                except:
                    lg.exc()
        backup_path = os.path.join(settings.getLocalBackupsDir(), latest_key_id)
        if not global_id.IsValidGlobalUser(latest_key_id):
            lg.warn('found incorrect folder name, not a customer: %s' % backup_path)
            continue
        if os.path.isdir(backup_path):
            bpio.traverse_dir_recursive(lambda r, s, n: visit(latest_key_id, r, s, n), backup_path)
        else:
            lg.warn('not a folder: %s' % backup_path)

    if _Debug:
        lg.out(_DebugLevel, 'backup_matrix.ReadLocalFiles %d files indexed' % _counter[0])
    if _LocalFilesNotifyCallback is not None:
        _LocalFilesNotifyCallback()


#------------------------------------------------------------------------------


def RemoteFileReport(backupID, blockNum, supplierNum, dataORparity, result, itemInfo):
    """
    Writes info for a single piece of data into "remote" matrix.

    May be called when you got an Ack packet from remote supplier after
    you sent him some Data packet.
    """
    blockNum = int(blockNum)
    supplierNum = int(supplierNum)
    customer_idurl = packetid.CustomerIDURL(backupID)
    if supplierNum > contactsdb.num_suppliers(customer_idurl=customer_idurl):
        if _Debug:
            lg.out(_DebugLevel, 'backup_matrix.RemoteFileReport got too big supplier number, possible this is an old packet')
        return
    if backupID not in remote_files():
        remote_files()[backupID] = {}
        lg.info('new remote entry for %s created in memory' % backupID)
    if blockNum not in remote_files()[backupID]:
        remote_files()[backupID][blockNum] = {
            'D': [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl),
            'P': [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl),
        }
    # save backed up block info into remote info structure, synchronize on hand info
    flag = 1 if result else 0
    if dataORparity == 'Data':
        remote_files()[backupID][blockNum]['D'][supplierNum] = flag
    elif dataORparity == 'Parity':
        remote_files()[backupID][blockNum]['P'][supplierNum] = flag
    else:
        lg.warn('incorrect backup ID: %s' % backupID)
    # if we know only N blocks stored on remote machine
    # but we uploaded N+1 block - remember that
    maxBlockNum = max(remote_max_block_numbers().get(backupID, -1), blockNum)
    remote_max_block_numbers()[backupID] = maxBlockNum
    full_remote_path = global_id.MakeGlobalID(path=itemInfo['name'], key_id=itemInfo['key_id'])
    full_remote_path_id = global_id.MakeGlobalID(path=itemInfo['path_id'], key_id=itemInfo['key_id'])
    _, percent, _, weakPercent = GetBackupRemoteStats(backupID)
    _, _, versionName = backupID.rpartition('/')
    snapshot = dict(
        backup_id=backupID,
        max_block=maxBlockNum,
        remote_path=full_remote_path,
        global_id=full_remote_path_id,
        type=itemInfo['type'],
        filesize=itemInfo['size'],
        size=itemInfo['versions'].get(versionName, [0, 0])[1],
        key_id=itemInfo['key_id'],
        delivered=misc.percent2string(percent),
        reliable=misc.percent2string(weakPercent),
    )
    listeners.push_snapshot('remote_version', snap_id=backupID, data=snapshot)
    if _Debug:
        lg.args(_DebugLevel, i=itemInfo['name'], b=backupID, s=supplierNum, n=blockNum, r=result)


def LocalFileReport(packetID=None, backupID=None, blockNum=None, supplierNum=None, dataORparity=None):
    """
    Writes info for a single piece of data into "local" matrix.

    You can use two forms:
    * pass ``packetID`` parameter only
    * pass all other parameters and do not use ``packetID``

    This is called when new local file created, for example during rebuilding process.
    """
    if packetID is not None:
        customer, remotePath, blockNum, supplierNum, dataORparity = packetid.Split(packetID)
        if remotePath is None:
            lg.warn('incorrect filename: ' + packetID)
            return
        backupID = packetid.MakeBackupID(customer, remotePath)
    else:
        blockNum = int(blockNum)
        supplierNum = int(supplierNum)
        dataORparity = dataORparity
        packetID = packetid.MakePacketID(backupID, blockNum, supplierNum, dataORparity)
    customer, filename = packetid.SplitPacketID(packetID)
    customer_idurl = global_id.GlobalUserToIDURL(customer)
    if dataORparity not in ['Data', 'Parity']:
        lg.warn('Data or Parity? ' + filename)
        return
    if supplierNum >= contactsdb.num_suppliers(customer_idurl=customer_idurl):
        if _Debug:
            lg.out(_DebugLevel, 'backup_matrix.LocalFileReport SKIP supplier position is invalid %d > %d for customer %s : %s' % (supplierNum, contactsdb.num_suppliers(), customer_idurl, filename))
        return
    supplier_idurl = contactsdb.supplier(supplierNum, customer_idurl=customer_idurl)
    if not supplier_idurl:
        if _Debug:
            lg.dbg(_DebugLevel, 'empty supplier at position %s for customer %s' % (supplierNum, customer_idurl))
        return
    localDest = os.path.join(settings.getLocalBackupsDir(), customer, filename)
    if backupID not in local_files():
        local_files()[backupID] = {}
    if blockNum not in local_files()[backupID]:
        local_files()[backupID][blockNum] = {
            'D': [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl),
            'P': [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl),
        }
    if not os.path.isfile(localDest):
        local_files()[backupID][blockNum][dataORparity[0]][supplierNum] = 0
        return
    local_files()[backupID][blockNum][dataORparity[0]][supplierNum] = 1
    if backupID not in local_max_block_numbers():
        local_max_block_numbers()[backupID] = -1
    if local_max_block_numbers()[backupID] < blockNum:
        local_max_block_numbers()[backupID] = blockNum
    if backupID not in local_backup_size():
        local_backup_size()[backupID] = 0
    try:
        local_backup_size()[backupID] += os.path.getsize(localDest)
    except:
        lg.exc()


def LocalBlockReport(backupID, blockNumber, result):
    """
    This updates "local" matrix - a several pieces corresponding to given block of data.
    """
    if result is None:
        lg.warn('result is None')
        return
    try:
        blockNum = int(blockNumber)
    except:
        lg.exc()
        return
    customer, _ = packetid.SplitPacketID(backupID)
    customer_idurl = global_id.GlobalUserToIDURL(customer)
    # repaint_flag = False
    if _Debug:
        lg.out(_DebugLevel, 'backup_matrix.LocalFileReport  in block %d at %s for %s' % (blockNumber, backupID, customer))
    num_suppliers = contactsdb.num_suppliers(customer_idurl=customer_idurl)
    for supplierNum in range(num_suppliers):
        supplier_idurl = contactsdb.supplier(supplierNum, customer_idurl=customer_idurl)
        if not supplier_idurl:
            lg.warn('unknown supplier_idurl supplierNum=%s for %s, customer_idurl=%s' % (supplierNum, backupID, customer_idurl))
            continue
        for dataORparity in ('Data', 'Parity'):
            packetID = packetid.MakePacketID(backupID, blockNum, supplierNum, dataORparity)
            local_file = os.path.join(settings.getLocalBackupsDir(), customer, packetID)
            if backupID not in local_files():
                local_files()[backupID] = {}
                # repaint_flag = True
                if _Debug:
                    lg.out(_DebugLevel, '    new local entry for %s created in memory' % backupID)
            if blockNum not in local_files()[backupID]:
                local_files()[backupID][blockNum] = {
                    'D': [
                        0,
                    ]*num_suppliers,
                    'P': [
                        0,
                    ]*num_suppliers,
                }
                # repaint_flag = True
            if not os.path.isfile(local_file):
                local_files()[backupID][blockNum][dataORparity[0]][supplierNum] = 0
                # repaint_flag = True
                continue
            local_files()[backupID][blockNum][dataORparity[0]][supplierNum] = 1
            if backupID not in local_backup_size():
                local_backup_size()[backupID] = 0
                # repaint_flag = True
            try:
                local_backup_size()[backupID] += os.path.getsize(local_file)
                # repaint_flag = True
            except:
                lg.exc()
            if _Debug:
                lg.out(_DebugLevel, '    OK, local backup size is %s and max block num is %s' % (local_backup_size()[backupID], local_max_block_numbers()[backupID]))
    if backupID not in local_max_block_numbers():
        local_max_block_numbers()[backupID] = -1
    if local_max_block_numbers()[backupID] < blockNum:
        local_max_block_numbers()[backupID] = blockNum


#------------------------------------------------------------------------------


def ScanMissingBlocks(backupID):
    """
    Finally here is some real logic.

    This will compare both matrixes to find missing pieces on remote
    suppliers. Should return a list of numbers of missed blocks for
    given backup.
    """
    if _Debug:
        lg.out(_DebugLevel, 'backup_matrix.ScanMissingBlocks for %s' % backupID)
    customer_idurl = packetid.CustomerIDURL(backupID)
    missingBlocks = set()
    localMaxBlockNum = local_max_block_numbers().get(backupID, -1)
    remoteMaxBlockNum = remote_max_block_numbers().get(backupID, -1)
    supplierActiveArray = GetActiveArray(customer_idurl=customer_idurl)

    if backupID not in remote_files():
        if backupID not in local_files():
            # we have no local and no remote info for this backup
            # no chance to do some rebuilds...
            # TODO: but how we get here ?!
            if _Debug:
                lg.out(_DebugLevel, '    no local and no remote info found !!!')
        else:
            # we have no remote info, but some local files exists
            # so let's try to sent all of them
            # need to scan all block numbers
            if _Debug:
                lg.out(_DebugLevel, '    no remote info but found local info, maxBlockNum=%d' % localMaxBlockNum)
            for blockNum in range(localMaxBlockNum + 1):
                # we check for Data and Parity packets
                localData = GetLocalDataArray(backupID, blockNum)
                localParity = GetLocalParityArray(backupID, blockNum)
                for supplierNum in range(len(supplierActiveArray)):
                    # if supplier is not alive we can not send to him
                    # so no need to scan for missing blocks
                    if supplierActiveArray[supplierNum] != 1:
                        continue
                    if localData[supplierNum] == 1:
                        missingBlocks.add(blockNum)
                    if localParity[supplierNum] == 1:
                        missingBlocks.add(blockNum)
    else:
        # now we have some remote info
        # we take max block number from local and remote
        maxBlockNum = max(remoteMaxBlockNum, localMaxBlockNum)
        if _Debug:
            lg.out(_DebugLevel, '    found remote info, maxBlockNum=%d' % maxBlockNum)
        # and increase by one because range(3) give us [0, 1, 2], but we want [0, 1, 2, 3]
        for blockNum in range(maxBlockNum + 1):
            # if we have few remote files, but many locals - we want to send all missed
            if blockNum not in remote_files()[backupID]:
                missingBlocks.add(blockNum)
                continue
            # take remote info for this block
            remoteData = GetRemoteDataArray(backupID, blockNum)
            remoteParity = GetRemoteParityArray(backupID, blockNum)
            # now check every our supplier for every block
            for supplierNum in range(len(supplierActiveArray)):
                # if supplier is not alive we can not send to him
                # so no need to scan for missing blocks
                if supplierActiveArray[supplierNum] != 1:
                    continue
                if supplierNum >= len(remoteData) or supplierNum >= len(remoteParity):
                    missingBlocks.add(blockNum)
                    continue
                if remoteData[supplierNum] != 1:  # -1 means missing
                    missingBlocks.add(blockNum)  # 0 - no info yet
                if remoteParity[supplierNum] != 1:  # 1 - file exist on remote supplier
                    missingBlocks.add(blockNum)

    if _Debug:
        lg.out(_DebugLevel, '    missingBlocks=%s' % missingBlocks)
    return list(missingBlocks)


def ScanBlocksToRemove(backupID, check_all_suppliers=True):
    """
    This method compare both matrixes and found pieces which is present on both
    sides.

    If remote supplier got that file it can be removed from the local
    HDD.
    """
    from bitdust.stream import io_throttle
    if _Debug:
        lg.out(_DebugLevel, 'backup_matrix.ScanBlocksToRemove for %r' % backupID)
    customer_idurl = packetid.CustomerIDURL(backupID)
    packets = []
    localMaxBlockNum = local_max_block_numbers().get(backupID, -1)
    if backupID not in remote_files() or backupID not in local_files():
        # no info about this backup yet - skip
        return packets
    for blockNum in range(localMaxBlockNum + 1):
        localArray = {'Data': GetLocalDataArray(backupID, blockNum), 'Parity': GetLocalParityArray(backupID, blockNum)}
        remoteArray = {'Data': GetRemoteDataArray(backupID, blockNum), 'Parity': GetRemoteParityArray(backupID, blockNum)}
        if (0 in remoteArray['Data']) or (0 in remoteArray['Parity']):
            # if some supplier do not have some data for that block - do not remove any local files for that block!
            # we do remove the local files only when we sure all suppliers got the all data pieces
            continue
        if (-1 in remoteArray['Data']) or (-1 in remoteArray['Parity']):
            # also if we do not have any info about this block for some supplier do not remove other local pieces
            continue
        for supplierNum in range(contactsdb.num_suppliers(customer_idurl=customer_idurl)):
            supplierIDURL = contactsdb.supplier(supplierNum, customer_idurl=customer_idurl)
            if not supplierIDURL:
                # supplier is unknown - skip
                continue
            for dataORparity in ['Data', 'Parity']:
                packetID = packetid.MakePacketID(backupID, blockNum, supplierNum, dataORparity)
                if io_throttle.HasPacketInSendQueue(supplierIDURL, packetID):
                    # if we do sending the packet at the moment - skip
                    continue
                if supplierNum >= len(localArray[dataORparity]):
                    lg.warn('wrong supplier %r position %d for customer %r' % (supplierIDURL, supplierNum, customer_idurl))
                    continue
                if localArray[dataORparity][supplierNum] == 1:
                    packets.append(packetID)
                    # lg.out(10, '    mark to remove %s, blockNum:%d remote:%s local:%s' % (packetID, blockNum, str(remoteArray), str(localArray)))


#                if check_all_suppliers:
#                    if localArray[dataORparity][supplierNum] == 1:
#                        packets.append(packetID)
#                else:
#                    if remoteArray[dataORparity][supplierNum] == 1 and localArray[dataORparity][supplierNum] == 1:
#                        packets.append(packetID)
    return packets


def ScanBlocksToSend(backupID, limit_per_supplier=None):
    """
    Opposite method - search for pieces which is not yet delivered to remote suppliers.
    """
    customer_idurl = packetid.CustomerIDURL(backupID)
    if id_url.is_some_empty(contactsdb.suppliers(customer_idurl=customer_idurl)):
        lg.warn('found empty suppliers, SKIP')
        return {}
    localMaxBlockNum = local_max_block_numbers().get(backupID, -1)
    supplierActiveArray = GetActiveArray(customer_idurl=customer_idurl)
    bySupplier = {}
    for supplierNum in range(len(supplierActiveArray)):
        bySupplier[supplierNum] = set()
    if backupID not in remote_files():
        # if _Debug:
        #     lg.out(_DebugLevel, 'backup_matrix.ScanBlocksToSend  backupID %r not found in remote files' % backupID)
        for blockNum in range(localMaxBlockNum + 1):
            localData = GetLocalDataArray(backupID, blockNum)
            localParity = GetLocalParityArray(backupID, blockNum)
            for supplierNum in range(len(supplierActiveArray)):
                if supplierActiveArray[supplierNum] != 1:
                    continue
                if supplierNum >= len(localData) or supplierNum >= len(localParity):
                    continue
                if localData[supplierNum] == 1:
                    bySupplier[supplierNum].add(packetid.MakePacketID(backupID, blockNum, supplierNum, 'Data'))
                if localParity[supplierNum] == 1:
                    bySupplier[supplierNum].add(packetid.MakePacketID(backupID, blockNum, supplierNum, 'Parity'))
                if limit_per_supplier:
                    if len(bySupplier[supplierNum]) > limit_per_supplier:
                        break
    else:
        # if _Debug:
        #     lg.out(_DebugLevel, 'backup_matrix.ScanBlocksToSend  backupID %r was found in remote files' % backupID)
        for blockNum in range(localMaxBlockNum + 1):
            remoteData = GetRemoteDataArray(backupID, blockNum)
            remoteParity = GetRemoteParityArray(backupID, blockNum)
            localData = GetLocalDataArray(backupID, blockNum)
            localParity = GetLocalParityArray(backupID, blockNum)
            for supplierNum in range(len(supplierActiveArray)):
                if supplierActiveArray[supplierNum] != 1:
                    continue
                if supplierNum >= len(remoteData) or supplierNum >= len(remoteParity):
                    continue
                if remoteData[supplierNum] != 1 and localData[supplierNum] == 1:
                    bySupplier[supplierNum].add(packetid.MakePacketID(backupID, blockNum, supplierNum, 'Data'))
                if remoteParity[supplierNum] != 1 and localParity[supplierNum] == 1:
                    bySupplier[supplierNum].add(packetid.MakePacketID(backupID, blockNum, supplierNum, 'Parity'))
                if limit_per_supplier:
                    if len(bySupplier[supplierNum]) > limit_per_supplier:
                        break
    return bySupplier


#------------------------------------------------------------------------------


def EraseBackupRemoteInfo(backupID):
    """
    Clear info only for given backup from "remote" matrix.
    """
    if backupID in remote_files():
        del remote_files()[backupID]  # remote_files().pop(backupID)
    if backupID in remote_max_block_numbers():
        del remote_max_block_numbers()[backupID]


def EraseBackupLocalInfo(backupID):
    """
    Clear info only for given backup from "local" matrix.
    """
    if backupID in local_files():
        del local_files()[backupID]  # local_files().pop(backupID)
    if backupID in local_max_block_numbers():
        del local_max_block_numbers()[backupID]
    if backupID in local_backup_size():
        del local_backup_size()[backupID]


#------------------------------------------------------------------------------


def ClearLocalInfo():
    """
    Completely clear the whole "local" matrix.
    """
    local_files().clear()
    local_max_block_numbers().clear()
    local_backup_size().clear()


def ClearRemoteInfo():
    """
    Completely clear the whole "remote" matrix, in other words - forget all info about suppliers files.
    """
    remote_files().clear()
    remote_max_block_numbers().clear()


def ClearSupplierRemoteInfo(supplierNum, customer_idurl=None, key_alias=None):
    """
    Clear only "single column" in the "remote" matrix corresponding to given
    supplier.
    """
    if not customer_idurl:
        customer_idurl = my_id.getIDURL()
    files = 0
    backups = 0
    for backupID in remote_files().keys():
        _key_alias, _customer_idurl = packetid.KeyAliasCustomer(backupID)
        if _customer_idurl == customer_idurl and (key_alias is None or key_alias == 'master' or _key_alias == key_alias):
            backups += 1
            for blockNum in remote_files()[backupID].keys():
                try:
                    if remote_files()[backupID][blockNum]['D'][supplierNum] == 1:
                        files += 1
                        remote_files()[backupID][blockNum]['D'][supplierNum] = 0
                except:
                    pass
                try:
                    if remote_files()[backupID][blockNum]['P'][supplierNum] == 1:
                        files += 1
                        remote_files()[backupID][blockNum]['P'][supplierNum] = 0
                except:
                    pass
    if _Debug:
        lg.args(_DebugLevel, files_cleaned=files, backups_cleaned=backups, supplier_pos=supplierNum, c=customer_idurl, k=key_alias)
    return files


#------------------------------------------------------------------------------


def GetBackupStats(backupID):
    """
    Collect needed info from "remote" matrix and create a detailed report about
    given backup.
    """
    customer_idurl = packetid.CustomerIDURL(backupID)
    if backupID not in remote_files():
        return 0, 0, [(0, 0)]*contactsdb.num_suppliers(customer_idurl=customer_idurl)
    percentPerSupplier = 100.0/contactsdb.num_suppliers(customer_idurl=customer_idurl)
    # ??? maxBlockNum = remote_max_block_numbers().get(backupID, -1)
    maxBlockNum = GetKnownMaxBlockNum(backupID)
    fileNumbers = [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl)
    totalNumberOfFiles = 0
    for blockNum in remote_files()[backupID].keys():
        for supplierNum in range(len(fileNumbers)):
            if supplierNum < contactsdb.num_suppliers(customer_idurl=customer_idurl):
                if supplierNum < len(remote_files()[backupID][blockNum]['D']):
                    if remote_files()[backupID][blockNum]['D'][supplierNum] == 1:
                        fileNumbers[supplierNum] += 1
                        totalNumberOfFiles += 1
                else:
                    lg.warn('wrong supplier position %d for customer %r in backup Data matrix, backupID=%r' % (supplierNum, customer_idurl, backupID))
                if supplierNum < len(remote_files()[backupID][blockNum]['P']):
                    if remote_files()[backupID][blockNum]['P'][supplierNum] == 1:
                        fileNumbers[supplierNum] += 1
                        totalNumberOfFiles += 1
                else:
                    lg.warn('wrong supplier position %d for customer %r in backup Parity matrix, backupID=%r' % (supplierNum, customer_idurl, backupID))
    statsArray = []
    for supplierNum in range(contactsdb.num_suppliers(customer_idurl=customer_idurl)):
        if maxBlockNum > -1:
            # 0.5 because we count both Parity and Data.
            percent = percentPerSupplier*0.5*fileNumbers[supplierNum]/(maxBlockNum + 1)
        else:
            percent = 0.0
        statsArray.append((percent, fileNumbers[supplierNum]))
    del fileNumbers
    return totalNumberOfFiles, maxBlockNum, statsArray


def GetBackupLocalStats(backupID):
    """
    Provide detailed info about local files for that backup. Return a tuple::

    (totalPercent, totalNumberOfFiles, totalSize, maxBlockNum,
    statsArray)
    """
    # ??? maxBlockNum = local_max_block_numbers().get(backupID, -1)
    customer_idurl = packetid.CustomerIDURL(backupID)
    maxBlockNum = GetKnownMaxBlockNum(backupID)
    if backupID not in local_files():
        return 0, 0, 0, maxBlockNum, [(0, 0)]*contactsdb.num_suppliers(customer_idurl=customer_idurl)
    percentPerSupplier = 100.0/contactsdb.num_suppliers(customer_idurl=customer_idurl)
    totalNumberOfFiles = 0
    fileNumbers = [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl)
    for blockNum in range(maxBlockNum + 1):
        if blockNum not in list(local_files()[backupID].keys()):
            continue


#    for blockNum in local_files()[backupID].keys():
        for supplierNum in range(len(fileNumbers)):
            if supplierNum < contactsdb.num_suppliers(customer_idurl=customer_idurl):
                if local_files()[backupID][blockNum]['D'][supplierNum] == 1:
                    fileNumbers[supplierNum] += 1
                    totalNumberOfFiles += 1
                if local_files()[backupID][blockNum]['P'][supplierNum] == 1:
                    fileNumbers[supplierNum] += 1
                    totalNumberOfFiles += 1
    statsArray = []
    for supplierNum in range(contactsdb.num_suppliers(customer_idurl=customer_idurl)):
        if maxBlockNum > -1:
            # 0.5 because we count both Parity and Data.
            percent = percentPerSupplier*0.5*fileNumbers[supplierNum]/(maxBlockNum + 1)
        else:
            percent = 0.0
        statsArray.append((percent, fileNumbers[supplierNum]))
    del fileNumbers
    totalPercent = 100.0*0.5*totalNumberOfFiles/((maxBlockNum + 1)*contactsdb.num_suppliers(customer_idurl=customer_idurl))
    return totalPercent, totalNumberOfFiles, local_backup_size().get(backupID, 0), maxBlockNum, statsArray


def GetBackupBlocksAndPercent(backupID):
    """
    Another method to get details about a backup.
    """
    if backupID not in remote_files():
        return 0, 0
    # get max block number
    # ??? maxBlockNum = remote_max_block_numbers().get(backupID, -1)
    maxBlockNum = GetKnownMaxBlockNum(backupID)
    if maxBlockNum == -1:
        return 0, 0
    customer_idurl = packetid.CustomerIDURL(backupID)
    # we count all remote files for this backup
    fileCounter = 0
    for blockNum in remote_files()[backupID].keys():
        for supplierNum in range(contactsdb.num_suppliers(customer_idurl=customer_idurl)):
            if remote_files()[backupID][blockNum]['D'][supplierNum] == 1:
                fileCounter += 1
            if remote_files()[backupID][blockNum]['P'][supplierNum] == 1:
                fileCounter += 1
    # +1 since zero based and *0.5 because Data and Parity
    return maxBlockNum + 1, 100.0*0.5*fileCounter/((maxBlockNum + 1)*contactsdb.num_suppliers(customer_idurl=customer_idurl))


def GetBackupRemoteStats(backupID, only_available_files=True):
    """
    This method found a most "weak" block of that backup, this is a block which
    pieces is kept by less suppliers from all other blocks.

    This is needed to detect the whole backup availability.
    Because if you loose at least one block of the backup - you will loose the whole backup.!

    The backup condition is equal to the "worst" block condition.
    Return a tuple::

      (blocks, percent, weakBlock, weakBlockPercent)
    """
    if backupID not in remote_files():
        return -1, 0, -1, 0
    # get max block number
    # ??? maxBlockNum = remote_max_block_numbers().get(backupID, -1)
    maxBlockNum = GetKnownMaxBlockNum(backupID)
    if maxBlockNum == -1:
        return -1, 0, -1, 0
    customer_idurl = packetid.CustomerIDURL(backupID)
    supplierCount = contactsdb.num_suppliers(customer_idurl=customer_idurl)
    fileCounter = 0
    weakBlockNum = -1
    lessSuppliers = supplierCount
    activeArray = GetActiveArray(customer_idurl=customer_idurl)
    # we count all remote files for this backup - scan all blocks
    for blockNum in range(maxBlockNum + 1):
        if blockNum not in list(remote_files()[backupID].keys()):
            lessSuppliers = 0
            weakBlockNum = blockNum
            continue
        goodSuppliers = supplierCount
        for supplierNum in range(supplierCount):
            if activeArray[supplierNum] != 1 and only_available_files:
                goodSuppliers -= 1
                continue
            try:
                remote_files()[backupID][blockNum]['D'][supplierNum]
                remote_files()[backupID][blockNum]['P'][supplierNum]
            except:
                goodSuppliers -= 1
                continue
            if remote_files()[backupID][blockNum]['D'][supplierNum] != 1 or remote_files()[backupID][blockNum]['P'][supplierNum] != 1:
                goodSuppliers -= 1
            if remote_files()[backupID][blockNum]['D'][supplierNum] == 1:
                fileCounter += 1
            if remote_files()[backupID][blockNum]['P'][supplierNum] == 1:
                fileCounter += 1
        if goodSuppliers < lessSuppliers:
            lessSuppliers = goodSuppliers
            weakBlockNum = blockNum
    # +1 since zero based and *0.5 because Data and Parity
    return (
        maxBlockNum + 1,
        100.0*0.5*fileCounter/((maxBlockNum + 1)*supplierCount),
        weakBlockNum,
        100.0*float(lessSuppliers)/float(supplierCount),
    )


def GetBackupRemoteArray(backupID):
    """
    Get info for given backup from "remote" matrix.
    """
    if backupID not in remote_files():
        return None
    maxBlockNum = GetKnownMaxBlockNum(backupID)
    if maxBlockNum == -1:
        return None
    return remote_files()[backupID]


def GetBackupLocalArray(backupID):
    """
    Get info for given backup from "local" matrix.
    """
    if backupID not in local_files():
        return None
    maxBlockNum = GetKnownMaxBlockNum(backupID)
    if maxBlockNum == -1:
        return None
    return local_files()[backupID]


def GetBackupIDs(remote=True, local=False, sorted_ids=False):
    """
    Return a list of backup IDs which is present in the matrixes.

    You can choose which matrix to use.
    """
    s = set()
    if remote:
        s.update(list(remote_files().keys()))
    if local:
        s.update(list(local_files().keys()))
    if sorted_ids:
        return misc.sorted_backup_ids(list(s))
    return list(s)


def GetKnownMaxBlockNum(backupID):
    """
    Return a maximum "known" block number for given backup.
    """
    return max(remote_max_block_numbers().get(backupID, -1), local_max_block_numbers().get(backupID, -1))


def GetLocalMatrix(backupID, blockNum):
    customer_idurl = packetid.CustomerIDURL(backupID)
    if backupID not in local_files():
        return {
            'D': [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl),
            'P': [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl),
        }
    if blockNum not in local_files()[backupID]:
        return {
            'D': [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl),
            'P': [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl),
        }
    return local_files()[backupID][blockNum]


def GetLocalDataArray(backupID, blockNum):
    """
    Get "local" info for a single block of given backup, this is for "Data"
    surface.
    """
    customer_idurl = packetid.CustomerIDURL(backupID)
    if backupID not in local_files():
        return [
            0,
        ]*contactsdb.num_suppliers(customer_idurl=customer_idurl)
    if blockNum not in local_files()[backupID]:
        return [
            0,
        ]*contactsdb.num_suppliers(customer_idurl=customer_idurl)
    return local_files()[backupID][blockNum]['D']


def GetLocalParityArray(backupID, blockNum):
    """
    Get "local" info for a single block of given backup, this is for "Parity"
    surface.
    """
    customer_idurl = packetid.CustomerIDURL(backupID)
    if backupID not in local_files():
        return [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl)
    if blockNum not in local_files()[backupID]:
        return [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl)
    return local_files()[backupID][blockNum]['P']


def GetRemoteMatrix(backupID, blockNum):
    customer_idurl = packetid.CustomerIDURL(backupID)
    if backupID not in remote_files():
        return {
            'D': [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl),
            'P': [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl),
        }
    if blockNum not in remote_files()[backupID]:
        return {
            'D': [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl),
            'P': [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl),
        }
    return remote_files()[backupID][blockNum]


def GetRemoteDataArray(backupID, blockNum):
    """
    Get "remote" info for a single block of given backup, this is for "Data"
    surface.
    """
    customer_idurl = packetid.CustomerIDURL(backupID)
    if backupID not in remote_files():
        return [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl)
    if blockNum not in remote_files()[backupID]:
        return [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl)
    return remote_files()[backupID][blockNum]['D']


def GetRemoteParityArray(backupID, blockNum):
    """
    Get "remote" info for a single block of given backup, this is for "Parity"
    surface.
    """
    customer_idurl = packetid.CustomerIDURL(backupID)
    if backupID not in remote_files():
        return [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl)
    if blockNum not in remote_files()[backupID]:
        return [0]*contactsdb.num_suppliers(customer_idurl=customer_idurl)
    return remote_files()[backupID][blockNum]['P']


def GetSupplierStats(supplierNum, customer_idurl=None):
    """
    Collect info from "remote" matrix about given supplier.
    """
    result = {}
    files = total = 0
    for backupID in remote_files().keys():
        if customer_idurl != packetid.CustomerIDURL(backupID):
            continue
        result[backupID] = {
            'data': 0,
            'parity': 0,
            'total': 0,
        }
        for blockNum in remote_files()[backupID].keys():
            if remote_files()[backupID][blockNum]['D'][supplierNum] == 1:
                result[backupID]['data'] += 1
                files += 1
            if remote_files()[backupID][blockNum]['P'][supplierNum] == 1:
                result[backupID]['parity'] += 1
                files += 1
            result[backupID]['total'] += 2
            total += 2
    return files, total, result


def GetWeakLocalBlock(backupID):
    """
    Scan all "local" blocks for given backup and find the most "weak" block.
    """
    customer_idurl = packetid.CustomerIDURL(backupID)
    supplierCount = contactsdb.num_suppliers(customer_idurl=customer_idurl)
    if backupID not in local_files():
        return -1, 0, supplierCount
    maxBlockNum = GetKnownMaxBlockNum(backupID)
    weakBlockNum = -1
    lessSuppliers = supplierCount
    for blockNum in range(maxBlockNum + 1):
        if blockNum not in list(local_files()[backupID].keys()):
            return blockNum, 0, supplierCount
        goodSuppliers = supplierCount
        for supplierNum in range(supplierCount):
            if local_files()[backupID][blockNum]['D'][supplierNum] != 1 or local_files()[backupID][blockNum]['P'][supplierNum] != 1:
                goodSuppliers -= 1
        if goodSuppliers < lessSuppliers:
            lessSuppliers = goodSuppliers
            weakBlockNum = blockNum
    return weakBlockNum, lessSuppliers, supplierCount


def GetWeakRemoteBlock(backupID):
    """
    Scan all "remote" blocks for given backup and find the most "weak" block -
    less suppliers keeps the data and stay online.
    """
    customer_idurl = packetid.CustomerIDURL(backupID)
    supplierCount = contactsdb.num_suppliers(customer_idurl=customer_idurl)
    if backupID not in remote_files():
        return -1, 0, supplierCount
    maxBlockNum = GetKnownMaxBlockNum(backupID)
    weakBlockNum = -1
    lessSuppliers = supplierCount
    activeArray = GetActiveArray(customer_idurl=customer_idurl)
    for blockNum in range(maxBlockNum + 1):
        if blockNum not in list(remote_files()[backupID].keys()):
            return blockNum, 0, supplierCount
        goodSuppliers = supplierCount
        for supplierNum in range(supplierCount):
            if activeArray[supplierNum] != 1:
                goodSuppliers -= 1
                continue
            if remote_files()[backupID][blockNum]['D'][supplierNum] != 1 or remote_files()[backupID][blockNum]['P'][supplierNum] != 1:
                goodSuppliers -= 1
        if goodSuppliers < lessSuppliers:
            lessSuppliers = goodSuppliers
            weakBlockNum = blockNum
    return weakBlockNum, lessSuppliers, supplierCount


#------------------------------------------------------------------------------


def SetBackupStatusNotifyCallback(callBack):
    """
    This is to catch in the GUI when some backups stats were changed.
    """
    global _BackupStatusNotifyCallback
    _BackupStatusNotifyCallback = callBack


def SetLocalFilesNotifyCallback(callback):
    """
    This is to catch in the GUI when some local files were changed.
    """
    global _LocalFilesNotifyCallback
    _LocalFilesNotifyCallback = callback


#------------------------------------------------------------------------------


def add_list_files_query_callback(customer_idurl, query_path, callback_method):
    global _ListFilesQueryCallbacks
    if (customer_idurl, query_path) not in _ListFilesQueryCallbacks:
        _ListFilesQueryCallbacks[(customer_idurl, query_path)] = []
    _ListFilesQueryCallbacks[(customer_idurl, query_path)].append(callback_method)
    if _Debug:
        lg.args(_DebugLevel, customer_idurl=customer_idurl, query_path=query_path, active_callbacks=len(_ListFilesQueryCallbacks))


def remove_list_files_query_callback(customer_idurl, query_path, callback_method):
    global _ListFilesQueryCallbacks
    if (customer_idurl, query_path) not in _ListFilesQueryCallbacks:
        raise Exception('callbacks for query %r : %r are not exist' % (
            customer_idurl,
            query_path,
        ))
    if callback_method not in _ListFilesQueryCallbacks[(customer_idurl, query_path)]:
        raise Exception('callback %r for query %r : %r not exist' % (
            callback_method,
            customer_idurl,
            query_path,
        ))
    _ListFilesQueryCallbacks[(customer_idurl, query_path)].remove(callback_method)
    if len(_ListFilesQueryCallbacks[(customer_idurl, query_path)]) == 0:
        _ListFilesQueryCallbacks.pop((customer_idurl, query_path))
    if _Debug:
        lg.args(_DebugLevel, customer_idurl=customer_idurl, query_path=query_path, active_callbacks=len(_ListFilesQueryCallbacks))


#------------------------------------------------------------------------------


def populate_remote_versions(key_id=None, remote_path=None, backup_id=None):
    if key_id:
        for backupID in GetBackupIDs(remote=True, local=False, sorted_ids=False):
            customer_idurl = packetid.CustomerIDURL(backupID)
            keyAlias, _, pathID, versionName = packetid.SplitBackupIDFull(backupID)
            if key_id == my_keys.make_key_id(alias=keyAlias, creator_idurl=customer_idurl):
                itemInfo = backup_fs.GetByID(pathID, iterID=backup_fs.fsID(customer_idurl, keyAlias))
                if itemInfo:
                    full_remote_path = global_id.MakeGlobalID(path=itemInfo.name(), key_id=itemInfo.key_id)
                    if full_remote_path == remote_path:
                        full_remote_path_id = global_id.MakeGlobalID(path=itemInfo.path_id, key_id=itemInfo.key_id)
                        _, percent, _, weakPercent = GetBackupRemoteStats(backupID)
                        snapshot = dict(
                            backup_id=backupID,
                            max_block=remote_max_block_numbers().get(backupID, -1),
                            remote_path=full_remote_path,
                            global_id=full_remote_path_id,
                            type=backup_fs.TYPES.get(itemInfo.type, 'UNKNOWN').lower(),
                            filesize=itemInfo.size,
                            size=max(0, itemInfo.get_version_size(versionName)),
                            key_id=itemInfo.key_id,
                            delivered=misc.percent2string(percent),
                            reliable=misc.percent2string(weakPercent),
                        )
                        listeners.push_snapshot('remote_version', snap_id=backupID, data=snapshot)
        return
    if remote_path:
        for backupID in GetBackupIDs(remote=True, local=False, sorted_ids=False):
            customer_idurl = packetid.CustomerIDURL(backupID)
            keyAlias, _, pathID, versionName = packetid.SplitBackupIDFull(backupID)
            itemInfo = backup_fs.GetByID(pathID, iterID=backup_fs.fsID(customer_idurl, keyAlias))
            if itemInfo:
                full_remote_path = global_id.MakeGlobalID(path=itemInfo.name(), key_id=itemInfo.key_id)
                if full_remote_path == remote_path:
                    full_remote_path_id = global_id.MakeGlobalID(path=itemInfo.path_id, key_id=itemInfo.key_id)
                    _, percent, _, weakPercent = GetBackupRemoteStats(backupID)
                    snapshot = dict(
                        backup_id=backupID,
                        max_block=remote_max_block_numbers().get(backupID, -1),
                        remote_path=full_remote_path,
                        global_id=full_remote_path_id,
                        type=backup_fs.TYPES.get(itemInfo.type, 'UNKNOWN').lower(),
                        filesize=itemInfo.size,
                        size=max(0, itemInfo.get_version_size(versionName)),
                        key_id=itemInfo.key_id,
                        delivered=misc.percent2string(percent),
                        reliable=misc.percent2string(weakPercent),
                    )
                    listeners.push_snapshot('remote_version', snap_id=backupID, data=snapshot)
        return
    if backup_id:
        backupID = backup_id
        customer_idurl = packetid.CustomerIDURL(backupID)
        keyAlias, _, pathID, versionName = packetid.SplitBackupIDFull(backupID)
        itemInfo = backup_fs.GetByID(pathID, iterID=backup_fs.fsID(customer_idurl, keyAlias))
        if itemInfo:
            full_remote_path = global_id.MakeGlobalID(path=itemInfo.name(), key_id=itemInfo.key_id)
            full_remote_path_id = global_id.MakeGlobalID(path=itemInfo.path_id, key_id=itemInfo.key_id)
            _, percent, _, weakPercent = GetBackupRemoteStats(backupID)
            snapshot = dict(
                backup_id=backupID,
                max_block=remote_max_block_numbers().get(backupID, -1),
                remote_path=full_remote_path,
                global_id=full_remote_path_id,
                type=backup_fs.TYPES.get(itemInfo.type, 'UNKNOWN').lower(),
                filesize=itemInfo.size,
                size=max(0, itemInfo.get_version_size(versionName)),
                key_id=itemInfo.key_id,
                delivered=misc.percent2string(percent),
                reliable=misc.percent2string(weakPercent),
            )
            listeners.push_snapshot('remote_version', snap_id=backupID, data=snapshot)
        return
    for backupID in GetBackupIDs(remote=True, local=False, sorted_ids=False):
        customer_idurl = packetid.CustomerIDURL(backupID)
        keyAlias, _, pathID, versionName = packetid.SplitBackupIDFull(backupID)
        itemInfo = backup_fs.GetByID(pathID, iterID=backup_fs.fsID(customer_idurl, keyAlias))
        if itemInfo:
            full_remote_path = global_id.MakeGlobalID(path=itemInfo.name(), key_id=itemInfo.key_id)
            full_remote_path_id = global_id.MakeGlobalID(path=itemInfo.path_id, key_id=itemInfo.key_id)
            _, percent, _, weakPercent = GetBackupRemoteStats(backupID)
            snapshot = dict(
                backup_id=backupID,
                max_block=remote_max_block_numbers().get(backupID, -1),
                remote_path=full_remote_path,
                global_id=full_remote_path_id,
                type=backup_fs.TYPES.get(itemInfo.type, 'UNKNOWN').lower(),
                filesize=itemInfo.size,
                size=max(0, itemInfo.get_version_size(versionName)),
                key_id=itemInfo.key_id,
                delivered=misc.percent2string(percent),
                reliable=misc.percent2string(weakPercent),
            )
            listeners.push_snapshot('remote_version', snap_id=backupID, data=snapshot)


def populate_remote_versions_deleted(backups2remove):
    for backupID in backups2remove:
        global_id, _, _ = backupID.rpartition('/')
        keyAlias, customerGlobalID, _, _ = packetid.SplitBackupIDFull(backupID)
        snapshot = dict(
            backup_id=backupID,
            global_id=global_id,
            key_id=my_keys.make_key_id(alias=keyAlias, creator_glob_id=customerGlobalID),
        )
        listeners.push_snapshot('remote_version', snap_id=backupID, deleted=True, data=snapshot)


#------------------------------------------------------------------------------

if __name__ == '__main__':
    init()
    # import pprint
    # pprint.pprint(GetBackupIds())



#!/usr/bin/env python
# restore_worker.py
#
# Copyright (C) 2008 Veselin Penev, http://bitdust.io
#
# This file (restore_worker.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: restore_worker
.. role:: red

BitDust restore_worker() Automat

EVENTS:
    * :red:`abort`
    * :red:`block-failed`
    * :red:`block-restored`
    * :red:`data-received`
    * :red:`data-receiving-started`
    * :red:`data-receiving-stopped`
    * :red:`init`
    * :red:`instant`
    * :red:`raid-done`
    * :red:`raid-failed`
    * :red:`request-failed`
    * :red:`request-finished`
    * :red:`timer-5sec`


At least for now we will work one block at a time, though packets in parallel.
We ask transport_control for all the data packets for a block then see if we
get them all or need to ask for some parity packets.  We do this till we have
gotten a block with the "LastBlock" flag set.  If we have tried several times
and not gotten data packets from a supplier we can flag him as suspect-bad
and start requesting a parity packet to cover him right away.

When we are missing a data packet we pick a parity packet where we have all the
other data packets for that parity so we can recover the missing data packet.
This network cost for this is just as low as if we read the data packet.
But most of the time we won't bother reading the parities.  Just uses up bandwidth.

We don't want to fire someone till
after we have finished a restore in case we have other problems and they might come
back to life and save us.  However, we might keep packets for a node we plan to fire.
This would make replacing them very easy.

At the "tar" level a user will have choice of full restore (to new empty system probably)
or to restore to another location on disk, or to just recover certain files.  However,
in this module we have to read block by block all of the blocks.

How do we restore if we lost everything?
Our public/private-key needs to be stored previously
on USB, in safe or deposit box (encrypted with passphrase or clear text).

The other thing we need is the backupIDs which we can get from our suppliers with the ListFiles command.

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from six.moves import range

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

#------------------------------------------------------------------------------

import os
import sys
import time

#------------------------------------------------------------------------------

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in restore.py')

from twisted.internet.defer import Deferred

#------------------------------------------------------------------------------

from bitdust.automats import automat

from bitdust.logs import lg

from bitdust.system import bpio
from bitdust.system import tmpfile

from bitdust.main import settings
from bitdust.main import events

from bitdust.lib import packetid
from bitdust.lib import strng

from bitdust.contacts import contactsdb

from bitdust.crypt import encrypted

from bitdust.p2p import online_status
from bitdust.p2p import propagate

from bitdust.stream import data_receiver
from bitdust.stream import io_throttle

from bitdust.raid import raid_worker
from bitdust.raid import eccmap

from bitdust.services import driver

from bitdust.userid import global_id
from bitdust.userid import id_url
from bitdust.userid import my_id

#------------------------------------------------------------------------------


class RestoreWorker(automat.Automat):

    """
    This class implements all the functionality of ``restore_worker()`` state machine.
    """

    timers = {
        'timer-5sec': (5.0, ['REQUESTED']),
    }

    def __init__(self, BackupID, OutputFile, KeyID=None, ecc_map=None, debug_level=_DebugLevel, log_events=False, log_transitions=_Debug, publish_events=False, **kwargs):
        """
        Builds `restore_worker()` state machine.
        """
        self.creator_id = my_id.getIDURL()
        self.backup_id = BackupID
        _parts = packetid.SplitBackupID(self.backup_id)
        self.customer_id = _parts[0]
        self.customer_idurl = global_id.GlobalUserToIDURL(self.customer_id)
        self.known_suppliers = []
        self.path_id = _parts[1]
        self.version = _parts[2]
        self.output_stream = OutputFile
        self.key_id = KeyID
        # is current active block - so when add 1 we get to first, which is 0
        self.block_number = -1
        self.bytes_written = 0
        self.OnHandData = []
        self.OnHandParity = []
        self.abort_flag = False
        self.done_flag = False
        self.EccMap = ecc_map or None
        self.max_errors = 0
        self.Started = time.time()
        self.LastAction = time.time()
        self.RequestFails = []
        self.block_requests = {}
        self.AlreadyRequestedCounts = {}
        # For anyone who wants to know when we finish
        self.MyDeferred = Deferred()
        self.packetInCallback = None
        self.blockRestoredCallback = None
        self.Attempts = 0

        super(RestoreWorker, self).__init__(name='restore_worker_%s' % self.version, state='AT_STARTUP', debug_level=debug_level, log_events=log_events, log_transitions=log_transitions, publish_events=publish_events, **kwargs)
        events.send('restore-started', data=dict(backup_id=self.backup_id))

    def set_packet_in_callback(self, cb):
        self.packetInCallback = cb

    def set_block_restored_callback(self, cb):
        self.blockRestoredCallback = cb

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        """
        Method to catch the moment when `restore_worker()` state were changed.
        """
        if event != 'instant':
            if newstate in ['REQUESTED', 'RECEIVING'] and oldstate not in ['REQUESTED', 'RECEIVING']:
                reactor.callLater(0.01, self.automat, 'instant')  # @UndefinedVariable

    def state_not_changed(self, curstate, event, *args, **kwargs):
        if event == 'data-received' and curstate in ['REQUESTED', 'RECEIVING']:
            reactor.callLater(0.01, self.automat, 'instant')  # @UndefinedVariable

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <http://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'REQUESTED'
                self.doInit(*args, **kwargs)
                self.doStartNewBlock(*args, **kwargs)
                self.doScanExistingPackets(*args, **kwargs)
                self.doRequestPackets(*args, **kwargs)
                self.Attempts = 1
        #---REQUESTED---
        elif self.state == 'REQUESTED':
            if event == 'data-receiving-started':
                self.state = 'RECEIVING'
            elif event == 'timer-5sec' and self.Attempts == 2:
                self.doPingOfflineSuppliers(*args, **kwargs)
            elif event == 'data-received':
                self.doSavePacket(*args, **kwargs)
            elif event == 'request-failed' and self.isStillCorrectable(*args, **kwargs) and self.Attempts < 3:
                self.doScanExistingPackets(*args, **kwargs)
                self.doRequestPackets(*args, **kwargs)
                self.Attempts += 1
            elif (event == 'abort' or
                  (event == 'request-failed' and
                   (not self.isStillCorrectable(*args, **kwargs) or self.Attempts >= 3))) or ((event == 'instant' or event == 'request-finished') and not self.isBlockReceiving(*args, **kwargs) and not self.isBlockFixable(*args, **kwargs)):
                self.state = 'FAILED'
                self.doDeleteAllRequests(*args, **kwargs)
                self.doRemoveTempFile(*args, **kwargs)
                self.doReportFailed(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif (event == 'instant' or event == 'request-finished') and self.isBlockFixable(*args, **kwargs):
                self.state = 'RAID'
                self.doReadRaid(*args, **kwargs)
        #---RECEIVING---
        elif self.state == 'RECEIVING':
            if event == 'data-receiving-stopped' and self.isStillCorrectable(*args, **kwargs):
                self.state = 'REQUESTED'
                self.doScanExistingPackets(*args, **kwargs)
                self.doRequestPackets(*args, **kwargs)
            elif event == 'data-received':
                self.doSavePacket(*args, **kwargs)
            elif (event == 'abort' or ((event == 'request-failed' or event == 'data-receiving-stopped') and
                                       not self.isStillCorrectable(*args, **kwargs))) or ((event == 'instant' or event == 'request-finished') and not self.isBlockReceiving(*args, **kwargs) and not self.isBlockFixable(*args, **kwargs)):
                self.state = 'FAILED'
                self.doDeleteAllRequests(*args, **kwargs)
                self.doRemoveTempFile(*args, **kwargs)
                self.doReportFailed(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif (event == 'instant' or event == 'request-finished') and self.isBlockFixable(*args, **kwargs):
                self.state = 'RAID'
                self.doReadRaid(*args, **kwargs)
        #---RAID---
        elif self.state == 'RAID':
            if event == 'raid-done':
                self.state = 'BLOCK'
                self.doRestoreBlock(*args, **kwargs)
            elif event == 'data-received':
                pass
            elif event == 'raid-failed' or event == 'abort':
                self.state = 'FAILED'
                self.doDeleteAllRequests(*args, **kwargs)
                self.doRemoveTempFile(*args, **kwargs)
                self.doReportFailed(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---BLOCK---
        elif self.state == 'BLOCK':
            if event == 'block-restored' and not self.isLastBlock(*args, **kwargs):
                self.state = 'REQUESTED'
                self.doWriteRestoredData(*args, **kwargs)
                self.doRemoveTempFile(*args, **kwargs)
                self.doStartNewBlock(*args, **kwargs)
                self.doScanExistingPackets(*args, **kwargs)
                self.doRequestPackets(*args, **kwargs)
                self.Attempts = 1
            elif event == 'block-restored' and self.isLastBlock(*args, **kwargs):
                self.state = 'DONE'
                self.doWriteRestoredData(*args, **kwargs)
                self.doDeleteAllRequests(*args, **kwargs)
                self.doRemoveTempFile(*args, **kwargs)
                self.doReportDone(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'data-received':
                pass
            elif event == 'block-failed' or event == 'abort':
                self.state = 'FAILED'
                self.doDeleteAllRequests(*args, **kwargs)
                self.doRemoveTempFile(*args, **kwargs)
                self.doReportFailed(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---DONE---
        elif self.state == 'DONE':
            pass
        #---FAILED---
        elif self.state == 'FAILED':
            pass
        return None

    def isLastBlock(self, *args, **kwargs):
        """
        Condition method.
        """
        NewBlock = args[0][0]
        return NewBlock.LastBlock

    def isStillCorrectable(self, *args, **kwargs):
        """
        Condition method.
        """
        if not self.EccMap:
            return False
        result = bool(len(self.RequestFails) <= self.max_errors)
        if _Debug:
            lg.out(_DebugLevel, 'restore_worker.isStillCorrectable max_errors=%d, fails=%d' % (self.max_errors, len(self.RequestFails)))
        return result

    def isBlockFixable(self, *args, **kwargs):
        """
        Condition method.
        """
        if not self.EccMap:
            return False
        result = self.EccMap.Fixable(self.OnHandData, self.OnHandParity)
        if _Debug:
            lg.args(_DebugLevel, block_number=self.block_number, result=result, OnHandData=self.OnHandData, OnHandParity=self.OnHandParity)
        return result

    def isBlockReceiving(self, *args, **kwargs):
        """
        Condition method.
        """
        block_results = list(self.block_requests.values())
        if _Debug:
            lg.args(_DebugLevel, block_results=block_results)
        return block_results.count(None) > 0

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self._do_block_rebuilding()
        self.known_suppliers = [_f for _f in contactsdb.suppliers(customer_idurl=self.customer_idurl) if _f]
        if not self.EccMap:
            if self.customer_idurl == my_id.getIDURL():
                self.EccMap = eccmap.Current()
                lg.info('ECC map %r set from local for my own suppliers' % self.EccMap)
        if not self.EccMap:
            known_eccmap_dict = {}
            for supplier_idurl in self.known_suppliers:
                known_ecc_map = contactsdb.get_supplier_meta_info(
                    supplier_idurl=supplier_idurl,
                    customer_idurl=self.customer_idurl,
                ).get('ecc_map', None)
                if known_ecc_map:
                    if known_ecc_map not in known_eccmap_dict:
                        known_eccmap_dict[known_ecc_map] = 0
                    known_eccmap_dict[known_ecc_map] += 1
            if known_eccmap_dict:
                all_known_eccmaps = list(known_eccmap_dict.items())
                all_known_eccmaps.sort(key=lambda i: i[1], reverse=True)
                self.EccMap = eccmap.eccmap(all_known_eccmaps[0][0])
                lg.info('ECC map %r recognized from suppliers meta info' % self.EccMap)
            else:
                known_ecc_map = None
                if driver.is_on('service_shared_data'):
                    from bitdust.access import shared_access_coordinator
                    active_share = shared_access_coordinator.get_active_share(self.key_id)
                    if active_share:
                        known_ecc_map = active_share.known_ecc_map
                if known_ecc_map:
                    self.EccMap = eccmap.eccmap(known_ecc_map)
                    lg.info('ECC map %r recognized from active share %r' % (self.EccMap, active_share))
                else:
                    num_suppliers = len(self.known_suppliers)
                    if num_suppliers not in eccmap.GetPossibleSuppliersCount():
                        num_suppliers = settings.DefaultDesiredSuppliers()
                    self.EccMap = eccmap.eccmap(eccmap.GetEccMapName(num_suppliers))
                    lg.warn('no meta info found, guessed ECC map %r from %d known suppliers' % (self.EccMap, len(self.known_suppliers)))
        self.max_errors = eccmap.GetCorrectableErrors(self.EccMap.NumSuppliers())
        if data_receiver.A():
            data_receiver.A().addStateChangedCallback(self._on_data_receiver_state_changed)

    def doStartNewBlock(self, *args, **kwargs):
        """
        Action method.
        """
        self.LastAction = time.time()
        self.block_number += 1
        if _Debug:
            lg.out(_DebugLevel, 'restore_worker.doStartNewBlock ' + str(self.block_number))
        self.OnHandData = [False]*self.EccMap.datasegments
        self.OnHandParity = [False]*self.EccMap.paritysegments
        self.RequestFails = []
        self.block_requests = {}
        self.AlreadyRequestedCounts = {}

    def doPingOfflineSuppliers(self, *args, **kwargs):
        """
        Action method.
        """
        l = []
        for supplier_idurl in contactsdb.suppliers(customer_idurl=self.customer_idurl):
            if online_status.isOnline(supplier_idurl):
                continue
            l.append(supplier_idurl)
        propagate.SendToIDs(l, wide=True)

    def doScanExistingPackets(self, *args, **kwargs):
        """
        Action method.
        """
        for SupplierNumber in range(self.EccMap.datasegments):
            PacketID = packetid.MakePacketID(self.backup_id, self.block_number, SupplierNumber, 'Data')
            customerID, remotePath = packetid.SplitPacketID(PacketID)
            self.OnHandData[SupplierNumber] = bool(os.path.exists(os.path.join(settings.getLocalBackupsDir(), customerID, remotePath)))
        for SupplierNumber in range(self.EccMap.paritysegments):
            PacketID = packetid.MakePacketID(self.backup_id, self.block_number, SupplierNumber, 'Parity')
            customerID, remotePath = packetid.SplitPacketID(PacketID)
            self.OnHandParity[SupplierNumber] = bool(os.path.exists(os.path.join(settings.getLocalBackupsDir(), customerID, remotePath)))

    def doRestoreBlock(self, *args, **kwargs):
        """
        Action method.
        """
        filename = args[0]
        blockbits = bpio.ReadBinaryFile(filename)
        if not blockbits:
            lg.warn('empty file %r' % filename)
            self.automat('block-failed')
            return
        try:
            splitindex = blockbits.index(b':')
            lengthstring = blockbits[0:splitindex]
            datalength = int(lengthstring)  # real length before raidmake/ECC
            blockdata = blockbits[splitindex + 1:splitindex + 1 + datalength]  # remove padding from raidmake/ECC
            newblock = encrypted.Unserialize(blockdata, decrypt_key=self.key_id)  # convert to object
        except:
            if _Debug:
                lg.exc('bad block: %r' % blockbits)
            else:
                lg.exc()
            self.automat('block-failed')
            return
        if not newblock:
            lg.warn('block read/unserialize failed from %d bytes of data' % len(blockbits))
            self.automat('block-failed')
            return
        self.automat('block-restored', (newblock, filename))

    def doRequestPackets(self, *args, **kwargs):
        """
        Action method.
        """
        self._do_check_run_requests()

    def doSavePacket(self, *args, **kwargs):
        """
        Action method.
        """
        if not args or not args[0]:
            raise Exception('no input found')
        NewPacket, PacketID = args[0]
        glob_path = global_id.NormalizeGlobalID(PacketID, detect_version=True)
        packetID = global_id.CanonicalID(PacketID)
        customer_id, _, _, _, SupplierNumber, dataORparity = packetid.SplitFull(packetID)
        if dataORparity == 'Data':
            self.OnHandData[SupplierNumber] = True
        elif dataORparity == 'Parity':
            self.OnHandParity[SupplierNumber] = True
        if not NewPacket:
            lg.warn('packet %r already exists locally' % packetID)
            return
        filename = os.path.join(settings.getLocalBackupsDir(), customer_id, glob_path['path'])
        dirpath = os.path.dirname(filename)
        if not os.path.exists(dirpath):
            try:
                bpio._dirs_make(dirpath)
            except:
                lg.exc()
        # either way the payload of packet is saved
        if not bpio.WriteBinaryFile(filename, NewPacket.Payload):
            lg.err('unable to write to %s' % filename)
            return
        if self.packetInCallback is not None:
            self.packetInCallback(self.backup_id, NewPacket)
        if _Debug:
            lg.out(_DebugLevel, 'restore_worker.doSavePacket %s saved to %s' % (packetID, filename))

    def doReadRaid(self, *args, **kwargs):
        """
        Action method.
        """
        alias = self.backup_id.split('$')[0]
        _, outfilename = tmpfile.make(
            'restore',
            extension='.raid',
            prefix=alias + '_' + str(self.block_number) + '_',
            close_fd=True,
        )
        inputpath = os.path.join(settings.getLocalBackupsDir(), self.customer_id, self.path_id)
        task_params = (outfilename, self.EccMap.name, self.version, self.block_number, inputpath)
        raid_worker.add_task('read', task_params, lambda cmd, params, result: self._on_block_restored(result, outfilename))

    def doRemoveTempFile(self, *args, **kwargs):
        """
        Action method.
        """
        if not args or not len(args) > 1:
            return
        filename = args[1]
        if filename:
            tmpfile.throw_out(filename, 'block restored')
        if settings.getBackupsKeepLocalCopies():
            return
        from bitdust.storage import backup_matrix
        from bitdust.storage import backup_rebuilder
        if not backup_rebuilder.ReadStoppedFlag():
            if backup_rebuilder.A().currentBackupID is not None:
                if backup_rebuilder.A().currentBackupID == self.backup_id:
                    if _Debug:
                        lg.out(_DebugLevel, 'restore_worker.doRemoveTempFile SKIP because rebuilding in process')
                    return
        count = 0
        for supplierNum in range(contactsdb.num_suppliers(customer_idurl=self.customer_idurl)):
            supplierIDURL = contactsdb.supplier(supplierNum, customer_idurl=self.customer_idurl)
            if not supplierIDURL:
                continue
            for dataORparity in ['Data', 'Parity']:
                packetID = packetid.MakePacketID(self.backup_id, self.block_number, supplierNum, dataORparity)
                customer, remotePath = packetid.SplitPacketID(packetID)
                filename = os.path.join(settings.getLocalBackupsDir(), customer, remotePath)
                if os.path.isfile(filename):
                    try:
                        os.remove(filename)
                    except:
                        lg.exc()
                        continue
                    count += 1
        backup_matrix.LocalBlockReport(self.backup_id, self.block_number, *args, **kwargs)
        if _Debug:
            lg.out(_DebugLevel, 'restore_worker.doRemoveTempFile %d files were removed' % count)

    def doWriteRestoredData(self, *args, **kwargs):
        """
        Action method.
        """
        NewBlock = args[0][0]
        data = NewBlock.Data()
        # Add to the file where all the data is going
        try:
            os.write(self.output_stream, data)
            self.bytes_written += len(data)
        except:
            lg.exc()
            # TODO Error handling...
            return
        if self.blockRestoredCallback is not None:
            self.blockRestoredCallback(self.backup_id, NewBlock)

    def doDeleteAllRequests(self, *args, **kwargs):
        """
        Action method.
        """
        io_throttle.DeleteBackupRequests(self.backup_id)

    def doReportDone(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'restore_worker.doReportDone')
        self.done_flag = True
        self.MyDeferred.callback('done')
        # events.send('restore-done', data=dict(backup_id=self.backup_id))

    def doReportFailed(self, *args, **kwargs):
        """
        Action method.
        """
        if args and len(args) > 0 and args[0] == 'abort':
            reason = 'abort'
        else:
            reason = 'failed'
        if _Debug:
            lg.out(_DebugLevel, 'restore_worker.doReportFailed : %s' % reason)
        self.done_flag = True
        self.MyDeferred.callback(reason)
        events.send('restore-failed', data=dict(
            backup_id=self.backup_id,
            block_number=self.block_number,
            args=args,
            reason=reason,
        ))

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        self._do_unblock_rebuilding()
        if data_receiver.A():
            data_receiver.A().removeStateChangedCallback(self._on_data_receiver_state_changed)
        self.OnHandData = None
        self.OnHandParity = None
        self.EccMap = None
        self.LastAction = None
        self.RequestFails = []
        self.AlreadyRequestedCounts = None
        self.block_requests = None
        self.MyDeferred = None
        self.output_stream = None
        self.destroy()

    def _do_block_rebuilding(self):
        from bitdust.storage import backup_rebuilder
        backup_rebuilder.BlockBackup(self.backup_id)
        io_throttle.DeleteBackupRequests(self.backup_id)

    def _do_unblock_rebuilding(self):
        from bitdust.storage import backup_rebuilder
        backup_rebuilder.UnBlockBackup(self.backup_id)

    def _do_check_run_requests(self):
        if _Debug:
            lg.out(_DebugLevel, 'restore_worker._do_check_run_requests for %s at block %d' % (self.backup_id, self.block_number))
        packetsToRequest = []
        for SupplierNumber in range(self.EccMap.datasegments):
            request_packet_id = packetid.MakePacketID(self.backup_id, self.block_number, SupplierNumber, 'Data')
            if self.OnHandData[SupplierNumber]:
                if _Debug:
                    lg.out(_DebugLevel, '        SKIP, OnHandData is True for supplier %d' % SupplierNumber)
                if request_packet_id not in self.block_requests:
                    self.block_requests[request_packet_id] = True
                continue
            if request_packet_id in self.block_requests:
                if _Debug:
                    lg.out(_DebugLevel, '        SKIP, request for packet %r already sent to IO queue for supplier %d' % (request_packet_id, SupplierNumber))
                continue
            SupplierID = contactsdb.supplier(SupplierNumber, customer_idurl=self.customer_idurl)
            if not SupplierID:
                lg.warn('unknown supplier at position %s' % SupplierNumber)
                continue
            if online_status.isOffline(SupplierID):
                if _Debug:
                    lg.out(_DebugLevel, '        SKIP, offline supplier: %s' % SupplierID)
                continue
            packetsToRequest.append((SupplierID, request_packet_id))
        for SupplierNumber in range(self.EccMap.paritysegments):
            request_packet_id = packetid.MakePacketID(self.backup_id, self.block_number, SupplierNumber, 'Parity')
            if self.OnHandParity[SupplierNumber]:
                if _Debug:
                    lg.out(_DebugLevel, '        SKIP, OnHandParity is True for supplier %d' % SupplierNumber)
                if request_packet_id not in self.block_requests:
                    self.block_requests[request_packet_id] = True
                continue
            if request_packet_id in self.block_requests:
                if _Debug:
                    lg.out(_DebugLevel, '        SKIP, request for packet %r already sent to IO queue for supplier %d' % (request_packet_id, SupplierNumber))
                continue
            SupplierID = contactsdb.supplier(SupplierNumber, customer_idurl=self.customer_idurl)
            if not SupplierID:
                lg.warn('unknown supplier at position %s' % SupplierNumber)
                continue
            if online_status.isOffline(SupplierID):
                if _Debug:
                    lg.out(_DebugLevel, '        SKIP, offline supplier: %s' % SupplierID)
                continue
            packetsToRequest.append((SupplierID, request_packet_id))
        requests_made = 0
        # already_requested = 0
        for SupplierID, packetID in packetsToRequest:
            if io_throttle.HasPacketInRequestQueue(SupplierID, packetID):
                # already_requested += 1
                # if packetID not in self.AlreadyRequestedCounts:
                #     self.AlreadyRequestedCounts[packetID] = 0
                # self.AlreadyRequestedCounts[packetID] += 1
                lg.warn('packet already in IO queue for supplier %s : %s' % (SupplierID, packetID))
                continue
            self.block_requests[packetID] = None
            if io_throttle.QueueRequestFile(
                callOnReceived=self._on_packet_request_result,
                creatorID=self.creator_id,
                packetID=packetID,
                ownerID=self.creator_id,  # self.customer_idurl,
                remoteID=SupplierID,
            ):
                requests_made += 1
            else:
                self.block_requests[packetID] = False
            if _Debug:
                lg.dbg(_DebugLevel, 'sent request %r to %r, other requests: %r' % (packetID, SupplierID, list(self.block_requests.values())))
        del packetsToRequest
        if requests_made:
            if _Debug:
                lg.out(_DebugLevel, '        requested %d packets for block %d' % (requests_made, self.block_number))
            return
        current_block_requests_results = list(self.block_requests.values())
        if _Debug:
            lg.args(_DebugLevel, current_results=current_block_requests_results)
        pending_count = current_block_requests_results.count(None)
        if pending_count > 0:
            if _Debug:
                lg.out(_DebugLevel, '        nothing for request, currently %d pending packets for block %d' % (pending_count, self.block_number))
            return
        failed_count = current_block_requests_results.count(False)
        if failed_count > self.max_errors:
            lg.err('all requests finished and %d packets failed, not possible to read data for block %d' % (failed_count, self.block_number))
            reactor.callLater(0, self.automat, 'request-failed', None)  # @UndefinedVariable
            return
        if _Debug:
            lg.out(_DebugLevel, '        all requests finished for block %d : %r' % (self.block_number, current_block_requests_results))
        reactor.callLater(0, self.automat, 'request-finished', None)  # @UndefinedVariable

    def _on_block_restored(self, restored_blocks, filename):
        if _Debug:
            lg.out(_DebugLevel, 'restore_worker._on_block_restored at %s with result: %s' % (filename, restored_blocks))
        if restored_blocks is None:
            self.automat('raid-failed', (None, filename))
        else:
            self.automat('raid-done', filename)

    def _on_packet_request_result(self, NewPacketOrPacketID, result):
        if self.block_requests is None:
            return
        if _Debug:
            lg.args(_DebugLevel, packet=NewPacketOrPacketID, result=result)
        packet_id = None
        if strng.is_string(NewPacketOrPacketID):
            packet_id = NewPacketOrPacketID
        else:
            packet_id = getattr(NewPacketOrPacketID, 'PacketID', None)
        if not packet_id:
            raise Exception('packet ID is unknown from %r' % NewPacketOrPacketID)
        if packet_id not in self.block_requests:
            resp = global_id.NormalizeGlobalID(packet_id)
            for req_packet_id in self.block_requests:
                req = global_id.NormalizeGlobalID(req_packet_id)
                if resp['version'] == req['version'] and resp['path'] == req['path']:
                    if resp['key_alias'] == req['key_alias'] and resp['user'] == req['user']:
                        if id_url.is_the_same(resp['idurl'], req['idurl']):
                            packet_id = req_packet_id
                            lg.warn('found matching packet request %r for rotated idurl %r' % (packet_id, resp['idurl']))
                            break
        if packet_id not in self.block_requests:
            if _Debug:
                lg.args(_DebugLevel, block_requests=self.block_requests)
            raise Exception('packet ID not registered')
        if result == 'in queue':
            if self.block_requests[packet_id] is not None:
                raise Exception('packet is still in IO queue, but already unregistered')
            lg.warn('packet already in the request queue: %r' % packet_id)
            return
        if result in ['received', 'exist']:
            self.block_requests[packet_id] = True
            if result == 'exist':
                # reactor.callLater(0, self.automat, 'data-received', (None, packet_id, ))  # @UndefinedVariable
                self.event('data-received', (None, packet_id))
            else:
                # reactor.callLater(0, self.automat, 'data-received', (NewPacketOrPacketID, packet_id, ))  # @UndefinedVariable
                self.event('data-received', (NewPacketOrPacketID, packet_id))
        else:
            self.block_requests[packet_id] = False
            self.RequestFails.append(packet_id)
            # reactor.callLater(0, self.automat, 'request-failed', packet_id)  # @UndefinedVariable
            self.event('request-failed', packet_id)

    def _on_data_receiver_state_changed(self, oldstate, newstate, event_string, *args, **kwargs):
        if newstate == 'RECEIVING' and oldstate != 'RECEIVING':
            self.automat('data-receiving-started', newstate)
        elif oldstate == 'RECEIVING' and newstate != 'RECEIVING':
            self.automat('data-receiving-stopped', newstate)



#!/usr/bin/python
# raid_worker.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (raid_worker.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: raid_worker.

.. role:: red

BitDust raid_worker Automat

.. raw:: html

    <a href="raid_worker.png" target="_blank">
    <img src="raid_worker.png" style="max-width:100%;">
    </a>

EVENTS:
    * :red:`init`
    * :red:`new-task`
    * :red:`process-started`
    * :red:`shutdown`
    * :red:`task-done`
    * :red:`task-started`
    * :red:`timer-1min`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 8

#------------------------------------------------------------------------------

import os
import sys
import time
import threading

from six.moves import range

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in raid_worker.py')

from twisted.internet import threads

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.system import bpio

from bitdust.raid import read
from bitdust.raid import make
from bitdust.raid import rebuild

#------------------------------------------------------------------------------

_MODULES = (
    'bitdust.raid.read',
    'bitdust.raid.make',
    'bitdust.raid.rebuild',
    'bitdust.raid.eccmap',
    'bitdust.raid.raidutils',
    'bitdust.logs.lg',
    'os',
    'sys',
    'copy',
    'array',
    'traceback',
    'six',
    'io',
)

_VALID_TASKS = {
    'make': (
        make.do_in_memory,
        (
            make.RoundupFile,
            make.ReadBinaryFile,
            make.WriteFile,
            make.ReadBinaryFileAsArray,
        ),
    ),
    'read': (
        read.raidread,
        (
            read.RebuildOne,
            read.ReadBinaryFile,
        ),
    ),
    'rebuild': (rebuild.rebuild, ()),
}

#------------------------------------------------------------------------------

_RaidWorker = None

#------------------------------------------------------------------------------


def add_task(cmd, params, callback):
    if _Debug:
        lg.args(_DebugLevel, cmd=cmd, params=params)
    A('new-task', (cmd, params, callback))


def cancel_task(cmd, first_parameter):
    if not A():
        if _Debug:
            lg.out(_DebugLevel, 'raid_worker.cancel_task SKIP _RaidWorker is not started')
        return False
    task_id = None
    found = False
    for t_id, t_cmd, t_params in A().tasks:
        if cmd == t_cmd and first_parameter == t_params[0]:
            try:
                A().tasks.remove(t_id, t_cmd, t_params)
                if _Debug:
                    lg.out(_DebugLevel, 'raid_worker.cancel_task found pending task %r, canceling %r' % (t_id, first_parameter))
            except:
                lg.warn('failed removing pending task %d, %s' % (t_id, first_parameter))
            found = True
            break
    for task_id, task_data in A().activetasks.items():
        t_proc, t_cmd, t_params = task_data
        if cmd == t_cmd and first_parameter == t_params[0]:
            if _Debug:
                lg.out(_DebugLevel, 'raid_worker.cancel_task found started task %r, aborting process %r' % (task_id, t_proc.tid))
            A().processor.cancel(t_proc.tid)
            found = True
            break
    if not found:
        lg.warn('task not found: %s %s' % (cmd, first_parameter))
        return False
    return True


#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _RaidWorker
    if _RaidWorker is None:
        # small workaround to not create a new instance during shutdown process
        if event is None or event != 'init':
            return None
        # set automat name and starting state here
        _RaidWorker = RaidWorker(
            name='raid_worker',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )
    if event is not None:
        _RaidWorker.automat(event, *args, **kwargs)
    return _RaidWorker


class RaidWorker(automat.Automat):
    """
    This class implements all the functionality of the ``raid_worker()`` state
    machine.
    """

    timers = {
        'timer-1min': (60, ['READY']),
    }

    def init(self):
        """
        Method to initialize additional variables and flags at creation of the
        state machine.
        """
        self.task_id = -1
        self.tasks = []
        self.activetasks = {}
        self.processor = None
        self.callbacks = {}

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'OFF'
                self.doInit(*args, **kwargs)
        #---OFF---
        elif self.state == 'OFF':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doKillProcess(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'new-task':
                self.doAddTask(*args, **kwargs)
                self.doStartProcess(*args, **kwargs)
            elif event == 'process-started' and self.isMoreTasks(*args, **kwargs):
                self.state = 'WORK'
                self.doStartTask(*args, **kwargs)
            elif event == 'process-started' and not self.isMoreTasks(*args, **kwargs):
                self.state = 'READY'
        #---READY---
        elif self.state == 'READY':
            if event == 'new-task':
                self.state = 'WORK'
                self.doAddTask(*args, **kwargs)
                self.doStartTask(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doKillProcess(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'timer-1min':
                self.state = 'OFF'
                self.doKillProcess(*args, **kwargs)
        #---WORK---
        elif self.state == 'WORK':
            if event == 'new-task':
                self.doAddTask(*args, **kwargs)
                self.doStartTask(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doReportTasksFailed(*args, **kwargs)
                self.doKillProcess(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'task-done' and self.isMoreTasks(*args, **kwargs):
                self.doReportTaskDone(*args, **kwargs)
                self.doPopTask(*args, **kwargs)
                self.doStartTask(*args, **kwargs)
            elif event == 'task-started' and self.isMoreTasks(*args, **kwargs):
                self.doStartTask(*args, **kwargs)
            elif event == 'task-done' and not self.isMoreActive(*args, **kwargs) and not self.isMoreTasks(*args, **kwargs):
                self.state = 'READY'
                self.doReportTaskDone(*args, **kwargs)
                self.doPopTask(*args, **kwargs)
            elif event == 'task-done' and self.isMoreActive(*args, **kwargs) and not self.isMoreTasks(*args, **kwargs):
                self.doReportTaskDone(*args, **kwargs)
                self.doPopTask(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        return None

    def isMoreTasks(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.tasks) > 0

    def isMoreActive(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.activetasks) > 1

    def doPopTask(self, *args, **kwargs):
        """
        Action method.
        """
        task_id, cmd, params, result = args[0]
        self.activetasks.pop(task_id)

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        reactor.addSystemEventTrigger('after', 'shutdown', self._kill_processor)  # @UndefinedVariable

    def doStartProcess(self, *args, **kwargs):
        """
        Action method.
        """
        self.processor = ThreadedRaidProcessor()

        # TODO: possible improvement in the future
        # On Android it is not possible to run a separate sub-process: the only possible way is to use threads
        # but on Windows and Linux we can use parallelp or multiprocessing libraries to utilize other CPU cores and
        # gain more performance

        # we do not want to use all CPU cores at once
        # need to keep at least one for all other operations
        # decided to use only half of CPUs at the moment
        # TODO: make an option in the software settings
        # ncpus = bpio.detect_number_of_cpu_cores()
        # if ncpus > 1:
        #     ncpus = int(ncpus/2.0)
        # if bpio.Android() or not config.conf().getBool('services/rebuilding/child-processes-enabled'):
        #     self.processor = ThreadedRaidProcessor()
        # else:
        #     os.environ['PYTHONUNBUFFERED'] = '1'
        #     from parallelp import pp
        #     self.processor = pp.Server(
        #         secret='bitdust',
        #         ncpus=ncpus,
        #         loglevel=lg.get_loging_level(_DebugLevel),
        #         logfile=settings.ParallelPLogFilename(),
        #     )
        # else:
        #     from bitdust.raid import worker
        #     self.processor = worker.Manager(ncpus=ncpus)

        self.automat('process-started')

    def doKillProcess(self, *args, **kwargs):
        """
        Action method.
        """
        self._kill_processor()
        self.processor = None
        self.automat('process-finished')

    def doAddTask(self, *args, **kwargs):
        """
        Action method.
        """
        cmd, params, callback = args[0]
        self.task_id += 1
        self.tasks.append((self.task_id, cmd, params))
        self.callbacks[self.task_id] = callback

    def doStartTask(self, *args, **kwargs):
        """
        Action method.
        """
        global _VALID_TASKS
        global _MODULES

        if len(self.activetasks) >= self.processor.get_ncpus():
            # lg.warn('SKIP active=%d cpus=%d' % (
            #         len(self.activetasks), self.processor.get_ncpus()))
            return

        try:
            task_id, cmd, params = self.tasks.pop(0)
            func, depfuncs = _VALID_TASKS[cmd]
        except:
            lg.exc()
            return

        proc = self.processor.submit(
            func,
            args=params,
            depfuncs=depfuncs,
            modules=_MODULES,
            callback=lambda result: self._job_done(task_id, cmd, params, result),  # error_callback=lambda err: self._job_failed(task_id, cmd, params, err),
        )

        self.activetasks[task_id] = (proc, cmd, params)
        if _Debug:
            lg.out(_DebugLevel, 'raid_worker.doStartTask job_id=%r active=%d cpus=%d %s' % (task_id, len(self.activetasks), self.processor.get_ncpus(), threading.currentThread().getName()))

        # reactor.callLater(0, self.automat, 'task-started', task_id)  # @UndefinedVariable
        self.automat('task-started', task_id)

    def doReportTaskDone(self, *args, **kwargs):
        """
        Action method.
        """
        task_id, cmd, params, result = args[0]
        cb = self.callbacks.pop(task_id)
        reactor.callLater(0, cb, cmd, params, result)  # @UndefinedVariable
        if result is not None:
            if _Debug:
                lg.out(_DebugLevel, 'raid_worker.doReportTaskDone callbacks: %d tasks: %d active: %d' % (len(self.callbacks), len(self.tasks), len(self.activetasks)))
        else:
            if _Debug:
                lg.out(_DebugLevel, 'raid_worker.doReportTaskDone result=None !!!!! callbacks: %d tasks: %d active: %d' % (len(self.callbacks), len(self.tasks), len(self.activetasks)))

    def doReportTasksFailed(self, *args, **kwargs):
        """
        Action method.
        """
        for i in range(len(self.tasks)):
            task_id, cmd, params = self.tasks[i]
            cb = self.callbacks.pop(task_id)
            reactor.callLater(0, cb, cmd, params, None)  # @UndefinedVariable
        for task_id, task_data in self.activetasks.items():
            cb = self.callbacks.pop(task_id)
            _, cmd, params = task_data
            reactor.callLater(0, cb, cmd, params, None)  # @UndefinedVariable

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        self.destroy()
        global _RaidWorker
        del _RaidWorker
        _RaidWorker = None

    def _job_done(self, task_id, cmd, params, result):
        if _Debug:
            lg.out(_DebugLevel, 'raid_worker._job_done %r : %r active:%r cmd=%r params=%r %s' % (task_id, result, list(self.activetasks.keys()), cmd, params, threading.currentThread().getName()))
        reactor.callFromThread(self.automat, 'task-done', (task_id, cmd, params, result))  # @UndefinedVariable

    # def _job_failed(self, task_id, cmd, params, err):
    #     lg.err('task %r FAILED : %r   active:%r cmd=%r params=%r' % (
    #         task_id, err, list(self.activetasks.keys()), cmd, params))
    #     self.automat('shutdown')

    def _kill_processor(self):
        if self.processor:
            self.processor.destroy()
            if _Debug:
                lg.out(_DebugLevel, 'raid_worker._kill_processor processor was destroyed')


#------------------------------------------------------------------------------


class RaidTask(object):
    def __init__(self, task_id, worker_method, worker_args=None, on_success=None, on_fail=None):
        self.task_id = task_id
        self.result = None
        self._on_success = on_success
        self._on_fail = on_fail
        self._worker_method = worker_method
        self._worker_args = worker_args or ()
        self._stopped = False
        self._bytes_processed = 0
        self._started = None

    def threshold_control(self, more_bytes):
        if self._stopped:
            return False
        self._bytes_processed += more_bytes
        time_running = time.time() - self._started
        if self._bytes_processed % 10000 == 0:
            if _Debug:
                lg.args(_DebugLevel, bytes_processed=self._bytes_processed, time_running=time_running)
            time.sleep(0.01)
        return True

    def run(self):
        args = self._worker_args + (self.threshold_control, )
        self.result = self._worker_method(*args)
        return self.result

    def start(self):
        self._started = time.time()
        d = threads.deferToThread(self.run)  # @UndefinedVariable
        if self._on_success:
            d.addCallback(self._on_success)
        if self._on_fail:
            d.addErrback(self._on_fail)

    def stop(self):
        self._stopped = True


#------------------------------------------------------------------------------


class RaidTaskInfo(object):
    def __init__(self, task_id):
        self.tid = task_id


#------------------------------------------------------------------------------


class ThreadedRaidProcessor(object):
    def __init__(self):
        self.latest_task_id = 0
        self.tasks = {}
        self.active_tasks = {}
        self.max_simultaneous_tasks = 1

    def cancel(self, task_id):
        if task_id not in self.active_tasks:
            lg.warn('can not cancel task %r, task was not found' % task_id)
            return
        self.active_tasks[task_id].stop()

    def destroy(self):
        for ts in self.active_tasks.values():
            ts.stop()

    def get_ncpus(self):
        return self.max_simultaneous_tasks

    def on_success(self, task_id, result, callback):
        ts = self.active_tasks.pop(task_id)
        if _Debug:
            lg.args(_DebugLevel, task_id=task_id, result=result, bytes_processed=ts._bytes_processed, active_tasks=list(self.active_tasks.keys()))
        reactor.callLater(0, callback, result)  # @UndefinedVariable
        reactor.callLater(0, self.process)  # @UndefinedVariable
        return None

    def on_fail(self, task_id, result, callback):
        ts = self.active_tasks.pop(task_id)
        if _Debug:
            lg.args(_DebugLevel, task_id=task_id, result=result, bytes_processed=ts._bytes_processed, active_tasks=list(self.active_tasks.keys()))
        reactor.callLater(0, callback, result)  # @UndefinedVariable
        reactor.callLater(0, self.process)  # @UndefinedVariable
        return None

    def process(self):
        if len(self.active_tasks) >= self.max_simultaneous_tasks:
            return False
        if not self.tasks:
            return False
        next_task_id = list(self.tasks.keys())[0]
        ts = self.tasks.pop(next_task_id)
        self.active_tasks[next_task_id] = ts
        ts.start()

    def submit(self, func, args=None, depfuncs=None, modules=None, callback=None):
        task_id = self.latest_task_id + 1
        if task_id in self.tasks:
            raise Exception('another RaidTask already exists with task_id=%r' % task_id)
        rt = RaidTask(
            task_id=task_id,
            worker_method=func,
            worker_args=args,
            on_success=lambda result: self.on_success(task_id, result, callback),
            on_fail=lambda result: self.on_fail(task_id, result, callback),
        )
        self.tasks[task_id] = rt
        self.latest_task_id = task_id
        if _Debug:
            lg.args(_DebugLevel, task_id=task_id, func=func, total_tasks=len(self.tasks))
        reactor.callLater(0, self.process)  # @UndefinedVariable
        return RaidTaskInfo(task_id)


#------------------------------------------------------------------------------


def _read_done(cmd, taskdata, result):
    lg.out(0, '_read_done %r %r %r' % (cmd, taskdata, result))
    A('shutdown')
    reactor.stop()  # @UndefinedVariable


def _make_done(cmd, taskdata, result):
    lg.out(0, '_make_done %r %r %r' % (cmd, taskdata, result))
    reactor.callLater(0.5, add_task, 'read', ('/tmp/destination.txt', 'ecc/18x18', 'F12345678', '5', '/tmp/raidtest'), _read_done)  # @UndefinedVariable


def main():
    import base64

    bpio.init()
    lg.set_debug_level(20)

    os.system('rm -rf /tmp/raidtest')
    os.system('mkdir -p /tmp/raidtest/F12345678')
    open('/tmp/source.txt', 'w').write(base64.b64encode(os.urandom(1000)).decode())

    reactor.callWhenRunning(A, 'init')  # @UndefinedVariable
    reactor.callLater(0.5, add_task, 'make', ('/tmp/source.txt', 'ecc/18x18', 'F12345678', '5', '/tmp/raidtest/F12345678'), _make_done)  # @UndefinedVariable

    reactor.run()  # @UndefinedVariable


if __name__ == '__main__':
    main()
